from typing import Any, Dict, List, Sequence, Tuple, cast, Optional, Callable, Union

import yaml
from dlt.common.time import ensure_pendulum_datetime_utc
from dlt.common.destination import PreparedTableSchema
from dlt.common.destination.utils import resolve_merge_strategy
from dlt.common.typing import TAnyDateTime, TypedDict

from dlt.common.schema.typing import (
    TSortOrder,
    TColumnProp,
)
from dlt.common.schema.utils import (
    get_columns_names_with_prop,
    get_first_column_name_with_prop,
    get_dedup_sort_tuple,
    get_validity_column_names,
    get_active_record_timestamp,
    is_nested_table,
)
from dlt.common.storages.load_storage import ParsedLoadJobFileName
from dlt.common.storages.load_package import load_package_state as current_load_package
from dlt.common.utils import uniq_id
from dlt.common.destination.capabilities import DestinationCapabilitiesContext
from dlt.destinations.exceptions import MergeDispositionException
from dlt.destinations.job_impl import FollowupJobRequestImpl
from dlt.destinations.sql_client import SqlClientBase
from dlt.common.destination.exceptions import DestinationTransientException


class SqlJobCreationException(DestinationTransientException):
    def __init__(
        self, original_exception: Exception, table_chain: Sequence[PreparedTableSchema]
    ) -> None:
        tables_str = yaml.dump(
            table_chain, allow_unicode=True, default_flow_style=False, sort_keys=False
        )
        super().__init__(
            f"Could not create SQLFollowupJob with exception {str(original_exception)}. Table"
            f" chain: {tables_str}"
        )


class SqlFollowupJob(FollowupJobRequestImpl):
    """Sql base job for jobs that rely on the whole table chain"""

    @classmethod
    def from_table_chain(
        cls,
        table_chain: Sequence[PreparedTableSchema],
        sql_client: SqlClientBase[Any],
    ) -> FollowupJobRequestImpl:
        """Generates a list of sql statements, that will be executed by the sql client when the job is executed in the loader.

        The `table_chain` contains a list of schemas of nested tables, ordered by the ancestry (the root of the tree is first on the list).
        """
        root_table = table_chain[0]
        file_info = ParsedLoadJobFileName(
            root_table["name"], ParsedLoadJobFileName.new_file_id(), 0, "sql"
        )

        try:
            # Remove line breaks from multiline statements and write one SQL statement per line in output file
            # to support clients that need to execute one statement at a time (i.e. snowflake)
            sql = [
                " ".join(stmt.splitlines()) for stmt in cls.generate_sql(table_chain, sql_client)
            ]
            job = cls(file_info.file_name())
            job._save_text_file("\n".join(sql))
        except Exception as e:
            raise SqlJobCreationException(e, table_chain) from e

        return job

    @classmethod
    def generate_sql(
        cls,
        table_chain: Sequence[PreparedTableSchema],
        sql_client: SqlClientBase[Any],
    ) -> List[str]:
        pass


class SqlStagingFollowupJob(SqlFollowupJob):
    """Generates a list of sql statements that copy the data from staging dataset into destination dataset."""

    @classmethod
    def _generate_insert_sql(
        cls,
        table_chain: Sequence[PreparedTableSchema],
        sql_client: SqlClientBase[Any],
        truncate_first: bool,
    ) -> List[str]:
        sql: List[str] = []
        for table in table_chain:
            with sql_client.with_staging_dataset():
                staging_table_name = sql_client.make_qualified_table_name(table["name"])
            table_name = sql_client.make_qualified_table_name(table["name"])
            columns = ", ".join(
                map(
                    sql_client.escape_column_name,
                    get_columns_names_with_prop(table, "name"),
                )
            )
            if truncate_first:
                sql.append(sql_client._truncate_table_sql(table_name))
            sql.append(
                f"INSERT INTO {table_name}({columns}) SELECT {columns} FROM {staging_table_name}"
            )
        return sql


class SqlStagingReplaceFollowupJob(SqlStagingFollowupJob):
    """Generates a list of sql statements that replace the data from staging dataset into destination dataset."""

    @classmethod
    def _generate_clone_sql(
        cls,
        table_chain: Sequence[PreparedTableSchema],
        sql_client: SqlClientBase[Any],
    ) -> List[str]:
        """Drop and clone the table for supported destinations"""
        sql: List[str] = []
        for table in table_chain:
            with sql_client.with_staging_dataset():
                staging_table_name = sql_client.make_qualified_table_name(table["name"])
            table_name = sql_client.make_qualified_table_name(table["name"])
            sql.append(f"DROP TABLE IF EXISTS {table_name}")
            # recreate destination table with data cloned from staging table
            sql.append(f"CREATE TABLE {table_name} CLONE {staging_table_name}")
        return sql

    @classmethod
    def generate_sql(
        cls,
        table_chain: Sequence[PreparedTableSchema],
        sql_client: SqlClientBase[Any],
    ) -> List[str]:
        root_table = table_chain[0]
        if (
            root_table["x-replace-strategy"] == "staging-optimized"  # type: ignore[typeddict-item]
            and sql_client.capabilities.supports_clone_table
        ):
            return cls._generate_clone_sql(table_chain, sql_client)

        return cls._generate_insert_sql(table_chain, sql_client, truncate_first=True)


class SqlStagingCopyFollowupJob(SqlStagingFollowupJob):
    """Generates a list of sql statements that copy the data from staging dataset into destination dataset."""

    @classmethod
    def generate_sql(
        cls,
        table_chain: Sequence[PreparedTableSchema],
        sql_client: SqlClientBase[Any],
    ) -> List[str]:
        return cls._generate_insert_sql(table_chain, sql_client, truncate_first=False)


class SqlMergeFollowupJob(SqlFollowupJob):
    """
    Generates a list of sql statements that merge the data from staging dataset into destination dataset.
    If no merge keys are discovered, falls back to append.
    """

    @classmethod
    def generate_sql(
        cls,
        table_chain: Sequence[PreparedTableSchema],
        sql_client: SqlClientBase[Any],
    ) -> List[str]:
        # resolve only root table
        root_table = table_chain[0]
        merge_strategy = resolve_merge_strategy(
            {root_table["name"]: root_table}, root_table, sql_client.capabilities
        )

        merge_sql = None
        if merge_strategy == "delete-insert":
            merge_sql = cls.gen_merge_sql(table_chain, sql_client)
        elif merge_strategy == "upsert":
            merge_sql = cls.gen_upsert_sql(table_chain, sql_client)
        elif merge_strategy == "scd2":
            merge_sql = cls.gen_scd2_sql(table_chain, sql_client)

        # prepend setup code
        return cls._gen_table_setup_clauses(table_chain, sql_client) + merge_sql

    @classmethod
    def _gen_table_setup_clauses(
        cls, table_chain: Sequence[PreparedTableSchema], sql_client: SqlClientBase[Any]
    ) -> List[str]:
        """Subclasses may override this method to generate additional sql statements to run before the merge"""
        return []

    @classmethod
    def _gen_key_table_clauses(
        cls, primary_keys: Sequence[str], merge_keys: Sequence[str]
    ) -> List[str]:
        """Generate sql clauses to select rows to delete via merge and primary key. Return select all clause if no keys defined."""
        clauses: List[str] = []
        if primary_keys or merge_keys:
            if primary_keys:
                clauses.append(
                    " AND ".join(["%s.%s = %s.%s" % ("{d}", c, "{s}", c) for c in primary_keys])
                )
            if merge_keys:
                clauses.append(
                    " AND ".join(["%s.%s = %s.%s" % ("{d}", c, "{s}", c) for c in merge_keys])
                )
        return clauses or ["1=1"]

    @classmethod
    def gen_key_table_clauses(
        cls,
        root_table_name: str,
        staging_root_table_name: str,
        key_clauses: Sequence[str],
        for_delete: bool,
    ) -> List[str]:
        """Generate sql clauses that may be used to select or delete rows in root table of destination dataset

        A list of clauses may be returned for engines that do not support OR in subqueries. Like BigQuery
        """
        return [
            f"FROM {root_table_name} as d WHERE EXISTS (SELECT 1 FROM {staging_root_table_name} as"
            f" s WHERE {' OR '.join([c.format(d='d',s='s') for c in key_clauses])})"
        ]

    @classmethod
    def gen_delete_temp_table_sql(
        cls,
        table_name: str,
        unique_column: str,
        key_table_clauses: Sequence[str],
        sql_client: SqlClientBase[Any],
    ) -> Tuple[List[str], str]:
        """Generate sql that creates delete temp table and inserts `unique_column` from root table for all records to delete. May return several statements.

        Returns temp table name for cases where special names are required like SQLServer.
        """
        sql: List[str] = []
        temp_table_name = cls._new_temp_table_name(table_name, "delete", sql_client)
        select_statement = f"SELECT d.{unique_column} {key_table_clauses[0]}"
        sql.append(cls._to_temp_table(select_statement, temp_table_name, unique_column))
        for clause in key_table_clauses[1:]:
            sql.append(f"INSERT INTO {temp_table_name} SELECT {unique_column} {clause}")
        return sql, temp_table_name

    @classmethod
    def gen_select_from_dedup_sql(
        cls,
        table_name: str,
        primary_keys: Sequence[str],
        columns: Sequence[str],
        dedup_sort: Tuple[str, TSortOrder] = None,
        condition: str = None,
        condition_columns: Sequence[str] = None,
        skip_dedup: bool = False,
    ) -> str:
        """Returns SELECT FROM SQL statement.

        The FROM clause in the SQL statement represents a deduplicated version
        of the `table_name` table.

        Expects column names provided in arguments to be escaped identifiers.

        Args:
            table_name: Name of the table that is selected from.
            primary_keys: A sequence of column names representing the primary
              key of the table. Is used to deduplicate the table.
            columns: Sequence of column names that will be selected from
              the table.
            dedup_sort: Name of a column and sort order tuple to sort the records by within a
              primary key. Values in the column are sorted in descending order,
              so the record with the highest value in `dedup_sort` remains
              after deduplication. No sorting is done if a None value is provided,
              leading to arbitrary deduplication.
            condition: String used as a WHERE clause in the SQL statement to
              filter records. The name of any column that is used in the
              condition but is not part of `columns` must be provided in the
              `condition_columns` argument. No filtering is done (aside from the
              deduplication) if a None value is provided.
            condition_columns: Sequence of names of columns used in the `condition`
              argument. These column names will be selected in the inner subquery
              to make them accessible to the outer WHERE clause. This argument
              should only be used in combination with the `condition` argument.
            skip_dedup: Skips deduplication if data declared deduplicated

        Returns:
            A string representing a SELECT FROM SQL statement where the FROM
            clause represents a deduplicated version of the `table_name` table.

            The returned value is used in two ways:
            1) To select the values for an INSERT INTO statement.
            2) To select the values for a temporary table used for inserts.
        """
        if condition is None:
            condition = "1 = 1"
        col_str = ", ".join(columns)
        inner_col_str = col_str
        if condition_columns is not None:
            inner_col_str += ", " + ", ".join(condition_columns)
        if skip_dedup:
            return f"SELECT {col_str} FROM {table_name} WHERE {condition}"
        else:
            order_by = cls.default_order_by()
            if dedup_sort is not None:
                order_by = f"{dedup_sort[0]} {dedup_sort[1].upper()}"
            return f"""
                SELECT {col_str}
                    FROM (
                        SELECT ROW_NUMBER() OVER (partition BY {", ".join(primary_keys)} ORDER BY {order_by}) AS _dlt_dedup_rn, {inner_col_str}
                        FROM {table_name}
                    ) AS _dlt_dedup_numbered WHERE _dlt_dedup_rn = 1 AND ({condition})

        """

    @classmethod
    def default_order_by(cls) -> str:
        return "(SELECT NULL)"

    @classmethod
    def gen_insert_temp_table_sql(
        cls,
        table_name: str,
        staging_root_table_name: str,
        sql_client: SqlClientBase[Any],
        primary_keys: Sequence[str],
        unique_column: str,
        dedup_sort: Tuple[str, TSortOrder] = None,
        condition: str = None,
        condition_columns: Sequence[str] = None,
        skip_dedup: bool = False,
    ) -> Tuple[List[str], str]:
        temp_table_name = cls._new_temp_table_name(table_name, "insert", sql_client)
        if len(primary_keys) > 0:
            # deduplicate
            select_sql = cls.gen_select_from_dedup_sql(
                staging_root_table_name,
                primary_keys,
                [unique_column],
                dedup_sort,
                condition,
                condition_columns,
                skip_dedup,
            )
        else:
            # don't deduplicate
            select_sql = f"SELECT {unique_column} FROM {staging_root_table_name} WHERE {condition}"
        return [cls._to_temp_table(select_sql, temp_table_name, unique_column)], temp_table_name

    @classmethod
    def gen_delete_from_sql(
        cls,
        table_name: str,
        unique_column: str,
        delete_temp_table_name: str,
        temp_table_column: str,
    ) -> str:
        """Generate DELETE FROM statement deleting the records found in the deletes temp table."""
        return f"""DELETE FROM {table_name}
            WHERE {unique_column} IN (
                SELECT * FROM {delete_temp_table_name}
            );
        """

    @classmethod
    def gen_concat_sql(cls, columns: Sequence[str]) -> str:
        return f"CONCAT({', '.join(columns)})"

    @classmethod
    def _shorten_table_name(cls, ident: str, sql_client: SqlClientBase[Any]) -> str:
        """Trims identifier to max length supported by sql_client. Used for dynamically constructed table names"""
        from dlt.common.normalizers.naming import NamingConvention

        return NamingConvention.shorten_identifier(
            ident, ident, sql_client.capabilities.max_identifier_length
        )

    @classmethod
    def _new_temp_table_name(cls, table_name: str, op: str, sql_client: SqlClientBase[Any]) -> str:
        return cls._shorten_table_name(f"{table_name}_{op}_{uniq_id()}", sql_client)

    @classmethod
    def _to_temp_table(cls, select_sql: str, temp_table_name: str, unique_column: str) -> str:
        """Generate sql that creates temp table from select statement. May return several statements.

        Args:
            select_sql: select statement to create temp table from
            temp_table_name: name of the temp table (unqualified)
            unique_column: column in the select list that is unique. used by Clickhouse only

        Returns:
            sql statement that inserts data from selects into temp table
        """
        return f"CREATE TEMPORARY TABLE {temp_table_name} AS {select_sql}"

    @classmethod
    def gen_update_table_prefix(cls, table_name: str) -> str:
        return f"UPDATE {table_name} SET"

    @classmethod
    def requires_temp_table_for_delete(cls) -> bool:
        """Whether a temporary table is required to delete records.

        Must be `True` for destinations that don't support correlated subqueries.
        """
        return False

    @classmethod
    def _escape_list(cls, list_: List[str], escape_id: Callable[[str], str]) -> List[str]:
        return list(map(escape_id, list_))

    @classmethod
    def _get_hard_delete_col_and_cond(
        cls,
        table: PreparedTableSchema,
        escape_id: Callable[[str], str],
        escape_lit: Callable[[Any], Any],
        invert: bool = False,
    ) -> Tuple[Optional[str], Optional[str]]:
        """Returns tuple of hard delete column name and SQL condition statement.

        Returns tuple of `None` values if no column has `hard_delete` hint.
        Condition statement can be used to filter deleted records.
        Set `invert=True` to filter non-deleted records instead.
        """

        col = get_first_column_name_with_prop(table, "hard_delete")
        if col is None:
            return (None, None)
        cond = f"{escape_id(col)} IS NOT NULL"
        if invert:
            cond = f"{escape_id(col)} IS NULL"
        if table["columns"][col]["data_type"] == "bool":
            if invert:
                cond += f" OR {escape_id(col)} = {escape_lit(False)}"
            else:
                cond = f"{escape_id(col)} = {escape_lit(True)}"
        return (col, cond)

    @classmethod
    def get_row_key_col(
        cls,
        table_chain: Sequence[PreparedTableSchema],
        table: PreparedTableSchema,
        dataset_name: str,
        staging_dataset_name: str,
    ) -> str:
        """Returns name of first column in `table` with `row_key` property. If not found first `unique` hint will be used.
        If no `unique` columns exist, will attempt to use a single primary key column.

        Returns:
            str: Name of the column to be used as row key

        Raises:
            MergeDispositionException: If no suitable column is found based on the search criteria
        """
        col = get_first_column_name_with_prop(table, "row_key")
        if col is not None:
            return col

        col = get_first_column_name_with_prop(table, "unique")
        if col is not None:
            return col

        # Try to use a single primary key column as a fallback
        primary_key_cols = get_columns_names_with_prop(table, "primary_key")
        if len(primary_key_cols) == 1:
            return primary_key_cols[0]
        elif len(primary_key_cols) > 1:
            raise MergeDispositionException(
                dataset_name,
                staging_dataset_name,
                [t["name"] for t in table_chain],
                f"Multiple primary key columns found in table `{table['name']}`. "
                "Cannot use as `row_key`.",
            )

        raise MergeDispositionException(
            dataset_name,
            staging_dataset_name,
            [t["name"] for t in table_chain],
            "No `row_key`, `unique`, or single primary key column (e.g. `_dlt_id`) "
            f"in table `{table['name']}`.",
        )

    @classmethod
    def get_root_key_col(
        cls,
        table_chain: Sequence[PreparedTableSchema],
        table: PreparedTableSchema,
        dataset_name: str,
        staging_dataset_name: str,
    ) -> str:
        """Returns name of first column in `table` with `root_key` property.

        Raises `MergeDispositionException` if no such column exists.
        """
        try:
            return cls._get_prop_col_or_raise(
                table,
                "root_key",
                MergeDispositionException(
                    dataset_name,
                    staging_dataset_name,
                    [t["name"] for t in table_chain],
                    f"No `root_key` column (e.g. `_dlt_root_id`) in table `{table['name']}`.",
                ),
            )
        except MergeDispositionException as merge_ex:
            # fallback to _dlt_parent_id is available if this is second nesting level
            if table["parent"] == table_chain[0]["name"]:
                return cls._get_prop_col_or_raise(
                    table,
                    "parent_key",
                    MergeDispositionException(
                        merge_ex.dataset_name,
                        merge_ex.staging_dataset_name,
                        merge_ex.tables,
                        merge_ex.reason
                        + "No `parent_key` column (e.g. `_dlt_parent_id`) in table"
                        f" `{table['name']}`.",
                    ),
                )
            else:
                raise

    @classmethod
    def _get_prop_col_or_raise(
        cls, table: PreparedTableSchema, prop: Union[TColumnProp, str], exception: Exception
    ) -> str:
        """Returns name of first column in `table` with `prop` property.

        Raises `exception` if no such column exists.
        """
        col = get_first_column_name_with_prop(table, prop)
        if col is None:
            raise exception
        return col

    @classmethod
    def gen_merge_sql(
        cls, table_chain: Sequence[PreparedTableSchema], sql_client: SqlClientBase[Any]
    ) -> List[str]:
        """Generates a list of sql statements that merge the data in staging dataset with the data in destination dataset.

        The `table_chain` contains a list schemas of a tables with row_key - parent_key nested reference, ordered by the ancestry (the root of the tree is first on the list).
        The root table is merged using primary_key and merge_key hints which can be compound and be both specified. In that case the OR clause is generated.
        The nested tables are merged based on propagated `root_key` which is a type of foreign key but always leading to a root table.

        First we store the root_keys of root table elements to be deleted in the temp table. Then we use the temp table to delete records from root and all netsed tables in the destination dataset.
        At the end we copy the data from the staging dataset into destination dataset.

        If a hard_delete column is specified, records flagged as deleted will be excluded from the copy into the destination dataset.
        If a dedup_sort column is specified in conjunction with a primary key, records will be sorted before deduplication, so the "latest" record remains.
        """
        sql: List[str] = []
        root_table = table_chain[0]

        escape_column_id = sql_client.escape_column_name
        escape_lit = sql_client.capabilities.escape_literal
        if escape_lit is None:
            escape_lit = DestinationCapabilitiesContext.generic_capabilities().escape_literal

        # get top level table full identifiers
        root_table_name, staging_root_table_name = sql_client.get_qualified_table_names(
            root_table["name"]
        )

        # get merge and primary keys from top level
        primary_keys = cls._escape_list(
            get_columns_names_with_prop(root_table, "primary_key"),
            escape_column_id,
        )
        merge_keys = cls._escape_list(
            get_columns_names_with_prop(root_table, "merge_key"),
            escape_column_id,
        )

        # if we do not have any merge keys to select from, we will fall back to a staged append, i.E.
        # just skip the delete part
        append_fallback = (len(primary_keys) + len(merge_keys)) == 0

        if not append_fallback:
            key_clauses = cls._gen_key_table_clauses(primary_keys, merge_keys)

            row_key_column: str = None
            root_key_column: str = None

            if len(table_chain) == 1 and not cls.requires_temp_table_for_delete():
                key_table_clauses = cls.gen_key_table_clauses(
                    root_table_name, staging_root_table_name, key_clauses, for_delete=True
                )
                # if no nested tables, just delete data from root table
                for clause in key_table_clauses:
                    sql.append(f"DELETE {clause}")
            else:
                key_table_clauses = cls.gen_key_table_clauses(
                    root_table_name, staging_root_table_name, key_clauses, for_delete=False
                )
                # use row_key or unique hint to create temp table with all identifiers to delete
                row_key_column = escape_column_id(
                    cls.get_row_key_col(
                        table_chain,
                        root_table,
                        sql_client.fully_qualified_dataset_name(),
                        sql_client.fully_qualified_dataset_name(staging=True),
                    )
                )
                create_delete_temp_table_sql, delete_temp_table_name = (
                    cls.gen_delete_temp_table_sql(
                        root_table["name"], row_key_column, key_table_clauses, sql_client
                    )
                )
                sql.extend(create_delete_temp_table_sql)

                # delete from nested tables first. This is important for databricks which does not support temporary tables,
                # but uses temporary views instead
                for table in table_chain[1:]:
                    table_name = sql_client.make_qualified_table_name(table["name"])
                    root_key_column = escape_column_id(
                        cls.get_root_key_col(
                            table_chain,
                            table,
                            sql_client.fully_qualified_dataset_name(),
                            sql_client.fully_qualified_dataset_name(staging=True),
                        )
                    )
                    sql.append(
                        cls.gen_delete_from_sql(
                            table_name, root_key_column, delete_temp_table_name, row_key_column
                        )
                    )

                # delete from root table now that nested tables have been processed
                sql.append(
                    cls.gen_delete_from_sql(
                        root_table_name, row_key_column, delete_temp_table_name, row_key_column
                    )
                )

        # get hard delete information
        hard_delete_col, not_deleted_cond = cls._get_hard_delete_col_and_cond(
            root_table,
            escape_column_id,
            escape_lit,
            invert=True,
        )

        # get dedup sort information
        dedup_sort = get_dedup_sort_tuple(root_table)
        if dedup_sort is not None:
            dedup_sort = (escape_column_id(dedup_sort[0]), dedup_sort[1])
        skip_dedup: bool = root_table.get("x-stage-data-deduplicated", False)  # type: ignore[assignment]

        insert_temp_table_name: str = None
        if len(table_chain) > 1:
            if len(primary_keys) > 0 or hard_delete_col is not None:
                # condition_columns = [hard_delete_col] if not_deleted_cond is not None else None
                condition_columns = None if hard_delete_col is None else [hard_delete_col]
                (
                    create_insert_temp_table_sql,
                    insert_temp_table_name,
                ) = cls.gen_insert_temp_table_sql(
                    root_table["name"],
                    staging_root_table_name,
                    sql_client,
                    primary_keys,
                    row_key_column,
                    dedup_sort,
                    not_deleted_cond,
                    condition_columns,
                    skip_dedup=skip_dedup,
                )
                sql.extend(create_insert_temp_table_sql)

        # insert from staging to dataset
        for table in table_chain:
            table_name, staging_table_name = sql_client.get_qualified_table_names(table["name"])

            insert_cond = not_deleted_cond if hard_delete_col is not None else "1 = 1"
            if (len(primary_keys) > 0 and len(table_chain) > 1) or (
                len(primary_keys) == 0
                and is_nested_table(table)  # nested table
                and hard_delete_col is not None
            ):
                uniq_column = root_key_column if is_nested_table(table) else row_key_column
                insert_cond = f"{uniq_column} IN (SELECT * FROM {insert_temp_table_name})"

            columns = list(map(escape_column_id, get_columns_names_with_prop(table, "name")))
            col_str = ", ".join(columns)
            select_sql = f"SELECT {col_str} FROM {staging_table_name} WHERE {insert_cond}"
            if len(primary_keys) > 0 and len(table_chain) == 1:
                # without nested tables we deduplicate inside the query instead of using a temp table
                select_sql = cls.gen_select_from_dedup_sql(
                    staging_table_name,
                    primary_keys,
                    columns,
                    dedup_sort,
                    insert_cond,
                    skip_dedup=skip_dedup,
                )

            sql.append(f"INSERT INTO {table_name}({col_str}) {select_sql}")
        return sql

    @classmethod
    def gen_upsert_merge_sql(
        cls,
        root_table_name: str,
        staging_root_table_name: str,
        primary_keys: Sequence[str],
        root_table_column_names: Sequence[str],
        hard_delete_col: Optional[str],
        deleted_cond: Optional[str],
    ) -> List[str]:
        """Generate MERGE statement for upsert on root table.

        Override for backends that don't support DELETE in MERGE (e.g., DuckLake).
        """
        sql: List[str] = []
        on_str = " AND ".join([f"d.{c} = s.{c}" for c in primary_keys])
        update_str = ", ".join([c + " = " + "s." + c for c in root_table_column_names])
        col_str = ", ".join(["{alias}" + c for c in root_table_column_names])
        delete_str = (
            "" if hard_delete_col is None else f"WHEN MATCHED AND s.{deleted_cond} THEN DELETE"
        )

        sql.append(f"""
            MERGE INTO {root_table_name} d USING {staging_root_table_name} s
            ON {on_str}
            {delete_str}
            WHEN MATCHED
                THEN UPDATE SET {update_str}
            WHEN NOT MATCHED
                THEN INSERT ({col_str.format(alias="")}) VALUES ({col_str.format(alias="s.")});
        """)
        return sql

    @classmethod
    def gen_upsert_sql(
        cls, table_chain: Sequence[PreparedTableSchema], sql_client: SqlClientBase[Any]
    ) -> List[str]:
        sql: List[str] = []
        root_table = table_chain[0]
        root_table_name, staging_root_table_name = sql_client.get_qualified_table_names(
            root_table["name"]
        )
        escape_column_id = sql_client.escape_column_name
        escape_lit = sql_client.capabilities.escape_literal
        if escape_lit is None:
            escape_lit = DestinationCapabilitiesContext.generic_capabilities().escape_literal

        # process table hints
        primary_keys = cls._escape_list(
            get_columns_names_with_prop(root_table, "primary_key"),
            escape_column_id,
        )
        hard_delete_col, deleted_cond = cls._get_hard_delete_col_and_cond(
            root_table,
            escape_column_id,
            escape_lit,
        )

        # generate merge statement for root table
        root_table_column_names = list(map(escape_column_id, root_table["columns"]))
        sql.extend(
            cls.gen_upsert_merge_sql(
                root_table_name,
                staging_root_table_name,
                primary_keys,
                root_table_column_names,
                hard_delete_col,
                deleted_cond,
            )
        )

        # generate statements for nested tables if they exist
        nested_tables = table_chain[1:]
        if nested_tables:
            root_row_key_column = escape_column_id(
                cls.get_row_key_col(
                    table_chain,
                    root_table,
                    sql_client.fully_qualified_dataset_name(),
                    sql_client.fully_qualified_dataset_name(staging=True),
                )
            )
            for table in nested_tables:
                nested_row_key_column = escape_column_id(
                    cls.get_row_key_col(
                        table_chain,
                        table,
                        sql_client.fully_qualified_dataset_name(),
                        sql_client.fully_qualified_dataset_name(staging=True),
                    )
                )
                root_key_column = escape_column_id(
                    cls.get_root_key_col(
                        table_chain,
                        table,
                        sql_client.fully_qualified_dataset_name(),
                        sql_client.fully_qualified_dataset_name(staging=True),
                    )
                )
                table_name, staging_table_name = sql_client.get_qualified_table_names(table["name"])

                # delete records for elements no longer in the list
                sql.append(f"""
                    DELETE FROM {table_name}
                    WHERE {root_key_column} IN (SELECT {root_row_key_column} FROM {staging_root_table_name})
                    AND {nested_row_key_column} NOT IN (SELECT {nested_row_key_column} FROM {staging_table_name});
                """)

                # insert records for new elements in the list
                table_column_names = list(map(escape_column_id, table["columns"]))
                update_str = ", ".join([c + " = " + "s." + c for c in table_column_names])
                col_str = ", ".join(["{alias}" + c for c in table_column_names])
                sql.append(f"""
                    MERGE INTO {table_name} d USING {staging_table_name} s
                    ON d.{nested_row_key_column} = s.{nested_row_key_column}
                    WHEN MATCHED
                        THEN UPDATE SET {update_str}
                    WHEN NOT MATCHED
                        THEN INSERT ({col_str.format(alias="")}) VALUES ({col_str.format(alias="s.")});
                """)

                # delete hard-deleted records
                if hard_delete_col is not None:
                    sql.append(f"""
                        DELETE FROM {table_name}
                        WHERE {root_key_column} IN (
                            SELECT {root_row_key_column}
                            FROM {staging_root_table_name}
                            WHERE {deleted_cond}
                        );
                    """)
        return sql

    @classmethod
    def gen_scd2_sql(
        cls, table_chain: Sequence[PreparedTableSchema], sql_client: SqlClientBase[Any]
    ) -> List[str]:
        """Generates SQL statements for the `scd2` merge strategy.

        The root table can be inserted into and updated.
        Updates only take place when a record retires (because there is a new version
        or it is deleted) and only affect the "valid to" column.
        Nested tables are insert-only.
        """
        sql: List[str] = []
        root_table = table_chain[0]
        root_table_name, staging_root_table_name = sql_client.get_qualified_table_names(
            root_table["name"]
        )

        # get column names
        caps = sql_client.capabilities
        escape_column_id = sql_client.escape_column_name
        from_, to = list(
            map(escape_column_id, get_validity_column_names(root_table))
        )  # validity columns
        hash_ = escape_column_id(
            get_first_column_name_with_prop(root_table, "x-row-version")
        )  # row hash column

        # define values for validity columns
        format_datetime_literal = caps.format_datetime_literal
        if format_datetime_literal is None:
            format_datetime_literal = (
                DestinationCapabilitiesContext.generic_capabilities().format_datetime_literal
            )

        _boundary_ts = cast(Optional[TAnyDateTime], root_table.get("x-boundary-timestamp"))
        boundary_ts: TAnyDateTime = (
            _boundary_ts
            if _boundary_ts is not None
            else current_load_package()["state"]["created_at"]
        )
        boundary_ts = ensure_pendulum_datetime_utc(boundary_ts)

        boundary_literal = format_datetime_literal(
            boundary_ts,
            caps.timestamp_precision,
        )

        active_record_timestamp = get_active_record_timestamp(root_table)
        if active_record_timestamp is None:
            active_record_literal = "NULL"
            is_active = f"{to} IS NULL"
        else:  # it's a datetime
            active_record_literal = format_datetime_literal(
                active_record_timestamp, caps.timestamp_precision
            )
            is_active = f"{to} = {active_record_literal}"

        # retire records:
        # - no `merge_key`: retire all absent records
        # - yes `merge_key`: retire those absent records whose `merge_key`
        # is present in staging data
        retire_sql = f"""
            {cls.gen_update_table_prefix(root_table_name)} {to} = {boundary_literal}
            WHERE {is_active}
            AND {hash_} NOT IN (SELECT {hash_} FROM {staging_root_table_name});
        """
        merge_keys = cls._escape_list(
            get_columns_names_with_prop(root_table, "merge_key"),
            escape_column_id,
        )
        if len(merge_keys) > 0:
            if len(merge_keys) == 1:
                key = merge_keys[0]
            else:
                key = cls.gen_concat_sql(merge_keys)  # compound key
            key_present = f"{key} IN (SELECT {key} FROM {staging_root_table_name})"
            retire_sql = retire_sql.rstrip()[:-1]  # remove semicolon
            retire_sql += f" AND {key_present}"
        sql.append(retire_sql)

        # insert new active records in root table
        columns = map(escape_column_id, list(root_table["columns"].keys()))
        col_str = ", ".join([c for c in columns if c not in (from_, to)])
        sql.append(f"""
            INSERT INTO {root_table_name} ({col_str}, {from_}, {to})
            SELECT {col_str}, {boundary_literal} AS {from_}, {active_record_literal} AS {to}
            FROM {staging_root_table_name} AS s
            WHERE {hash_} NOT IN (SELECT {hash_} FROM {root_table_name} WHERE {is_active});
        """)

        # insert list elements for new active records in nested tables
        nested_tables = table_chain[1:]
        if nested_tables:
            # TODO: - based on deterministic nested hashes (OK)
            # - if row hash changes all is right
            # - if it does not we only capture new records, while we should replace existing with those in stage
            # - this write disposition is way more similar to regular merge (how root tables are handled is different, other tables handled same)
            for table in nested_tables:
                row_key_column = escape_column_id(
                    cls.get_row_key_col(
                        table_chain,
                        table,
                        sql_client.fully_qualified_dataset_name(),
                        sql_client.fully_qualified_dataset_name(staging=True),
                    )
                )
                table_name, staging_table_name = sql_client.get_qualified_table_names(table["name"])
                sql.append(f"""
                    INSERT INTO {table_name}
                    SELECT *
                    FROM {staging_table_name}
                    WHERE {row_key_column} NOT IN (SELECT {row_key_column} FROM {table_name});
                """)

        return sql



import os
from datetime import datetime  # noqa: I251
from typing import (
    Generic,
    ClassVar,
    Any,
    Optional,
    Type,
    Dict,
    Union,
    Literal,
    Tuple,
)

import inspect
from functools import wraps

from dlt.common import logger
from dlt.common.exceptions import MissingDependencyException, ValueErrorWithKnownValues
from dlt.common.pendulum import pendulum
from dlt.common.jsonpath import compile_path, extract_simple_field_name
from dlt.common.typing import (
    TDataItem,
    TDataItems,
    TFun,
    TSortOrder,
    get_args,
    extract_inner_type,
    get_generic_type_argument_from_instance,
    is_optional_type,
    is_subclass,
    TColumnNames,
    TypedDict,
    resolve_single_annotation,
)
from dlt.common.configuration import configspec, ConfigurationValueError
from dlt.common.configuration.specs import BaseConfiguration
from dlt.common.data_types.type_helpers import (
    coerce_from_date_types,
    coerce_value,
    py_type_to_sc_type,
)
from dlt.common.data_writers.writers import count_rows_in_items
from dlt.extract.exceptions import IncrementalUnboundError
from dlt.extract.incremental.exceptions import (
    IncrementalCursorPathMissing,
    IncrementalPrimaryKeyMissing,
)
from dlt.common.incremental.typing import (
    IncrementalColumnState,
    TCursorValue,
    LastValueFunc,
    OnCursorValueMissing,
    IncrementalArgs,
    TIncrementalRange,
)
from dlt.extract.items import SupportsPipe, TTableHintTemplate
from dlt.extract.items_transform import BaseItemTransform, ItemTransform
from dlt.extract.state import resource_state
from dlt.extract.incremental.transform import (
    JsonIncremental,
    ArrowIncremental,
    IncrementalTransform,
)
from dlt.extract.incremental.lag import apply_lag

try:
    from dlt.common.libs.pyarrow import is_arrow_item
except MissingDependencyException:
    is_arrow_item = lambda item: False

try:
    from dlt.common.libs.pandas import pandas
except MissingDependencyException:
    pandas = None


class IncrementalCustomMetrics(TypedDict, total=False):
    unfiltered_items_count: int
    unfiltered_batches_count: int
    initial_unique_hashes_count: int
    final_unique_hashes_count: int


@configspec
class Incremental(
    ItemTransform[TDataItem, IncrementalCustomMetrics], BaseConfiguration, Generic[TCursorValue]
):
    """Adds incremental extraction for a resource by storing a cursor value in persistent state.

    The cursor could for example be a timestamp for when the record was created and you can use this to load only
    new records created since the last run of the pipeline.

    To use this the resource function should have an argument either type annotated with `Incremental` or a default `Incremental` instance.
    For example:

    >>> @dlt.resource(primary_key='id')
    >>> def some_data(created_at=dlt.sources.incremental('created_at', '2023-01-01T00:00:00Z'):
    >>>    yield from request_data(created_after=created_at.last_value)

    When the resource has a `primary_key` specified this is used to deduplicate overlapping items with the same cursor value.

    Alternatively you can use this class as transform step and add it to any resource. For example:
    >>> @dlt.resource
    >>> def some_data():
    >>>     last_value = dlt.sources.incremental.from_existing_state("some_data", "item.ts")
    >>>     ...
    >>>
    >>> r = some_data().add_step(dlt.sources.incremental("item.ts", initial_value=now, primary_key="delta"))
    >>> info = p.run(r, destination="duckdb")

    Args:
        cursor_path: The name or a JSON path to a cursor field. Uses the same names of fields as in your JSON document, before they are normalized to store in the database.
        initial_value: Optional value used for `last_value` when no state is available, e.g. on the first run of the pipeline. If not provided `last_value` will be `None` on the first run.
        last_value_func: Callable used to determine which cursor value to save in state. It is called with a list of the stored state value and all cursor vals from currently processing items. Default is `max`
        primary_key: Optional primary key used to deduplicate data. If not provided, a primary key defined by the resource will be used. Pass a tuple to define a compound key. Pass empty tuple to disable unique checks
        end_value: Optional value used to load a limited range of records between `initial_value` and `end_value`.
            Use in conjunction with `initial_value`, e.g. load records from given month `incremental(initial_value="2022-01-01T00:00:00Z", end_value="2022-02-01T00:00:00Z")`
            Note, when this is set the incremental filtering is stateless and `initial_value` always supersedes any previous incremental value in state.
        row_order: Declares that data source returns rows in descending (desc) or ascending (asc) order as defined by `last_value_func`. If row order is know, Incremental class
                    is able to stop requesting new rows by closing pipe generator. This prevents getting more data from the source. Defaults to None, which means that
                    row order is not known.
        allow_external_schedulers: If set to True, allows dlt to look for external schedulers from which it will take "initial_value" and "end_value" resulting in loading only
            specified range of data. Currently Airflow scheduler is detected: "data_interval_start" and "data_interval_end" are taken from the context and passed Incremental class.
            The values passed explicitly to Incremental will be ignored.
            Note that if logical "end date" is present then also "end_value" will be set which means that resource state is not used and exactly this range of date will be loaded
        on_cursor_value_missing: Specify what happens when the cursor_path does not exist in a record or a record has `None` at the cursor_path: raise, include, exclude
        lag: Optional value used to define a lag or attribution window. For datetime cursors, this is interpreted as seconds. For other types, it uses the + or - operator depending on the last_value_func.
        range_start: Decide whether the incremental filtering range is `open` or `closed` on the start value side. Default is `closed`.
            Setting this to `open` means that items with the same cursor value as the last value from the previous run (or `initial_value`) are excluded from the result.
            The `open` range disables deduplication logic so it can serve as an optimization when you know cursors don't overlap between pipeline runs.
        range_end: Decide whether the incremental filtering range is `open` or `closed` on the end value side. Default is `open` (exact `end_value` is excluded).
            Setting this to `closed` means that items with the exact same cursor value as the `end_value` are included in the result.
    """

    # this is config/dataclass so declare members
    cursor_path: str = None
    # TODO: Support typevar here
    initial_value: Optional[Any] = None
    end_value: Optional[Any] = None
    row_order: Optional[TSortOrder] = None
    allow_external_schedulers: bool = False
    on_cursor_value_missing: OnCursorValueMissing = "raise"
    lag: Optional[float] = None
    duplicate_cursor_warning_threshold: ClassVar[int] = 200
    range_start: TIncrementalRange = "closed"
    range_end: TIncrementalRange = "open"

    # incremental acting as empty
    EMPTY: ClassVar["Incremental[Any]"] = None
    placement_affinity: ClassVar[float] = 1  # stick to end

    def __init__(
        self,
        cursor_path: str = None,
        initial_value: Optional[TCursorValue] = None,
        last_value_func: Optional[Union[LastValueFunc[TCursorValue], Literal["min", "max"]]] = max,
        primary_key: Optional[TTableHintTemplate[TColumnNames]] = None,
        end_value: Optional[TCursorValue] = None,
        row_order: Optional[TSortOrder] = None,
        allow_external_schedulers: bool = False,
        on_cursor_value_missing: OnCursorValueMissing = "raise",
        lag: Optional[float] = None,
        range_start: TIncrementalRange = "closed",
        range_end: TIncrementalRange = "open",
    ) -> None:
        # make sure that path is valid
        if cursor_path:
            compile_path(cursor_path)
        self.cursor_path = cursor_path
        if isinstance(last_value_func, str):
            if last_value_func == "min":
                last_value_func = min
            elif last_value_func == "max":
                last_value_func = max
            else:
                raise ValueError(
                    f"Unknown `{last_value_func=:}` passed as string. Provide a"
                    " callable to use a custom function."
                )
        self.last_value_func = last_value_func
        self.initial_value = initial_value
        """Initial value of last_value"""
        self.end_value = end_value
        self.start_value: Any = initial_value
        """Value of last_value at the beginning of current pipeline run"""
        self.resource_name: Optional[str] = None
        # TODO: deprecate primary_key, use deduplication_key
        self._primary_key: Optional[TTableHintTemplate[TColumnNames]] = primary_key
        self.row_order = row_order
        self.allow_external_schedulers = allow_external_schedulers
        if on_cursor_value_missing not in ["raise", "include", "exclude"]:
            raise ValueErrorWithKnownValues(
                "on_cursor_value_missing", on_cursor_value_missing, ["raise", "include", "exclude"]
            )

        self.on_cursor_value_missing = on_cursor_value_missing

        self._cached_state: IncrementalColumnState = None
        """State dictionary cached on first access"""

        self.lag = lag
        super().__init__(lambda x: x)  # TODO:

        self.end_out_of_range: bool = False
        """Becomes true on the first item that is out of range of `end_value`. I.e. when using `max` function this means a value that is equal or higher"""
        self.start_out_of_range: bool = False
        """Becomes true on the first item that is out of range of `start_value`. I.e. when using `max` this is a value that is lower than `start_value`"""

        self._transformers: Dict[Type[IncrementalTransform], IncrementalTransform] = {}
        self._bound_pipe: SupportsPipe = None
        """Bound pipe"""
        self.range_start = range_start
        self.range_end = range_end
        self._custom_metrics: IncrementalCustomMetrics = {
            "unfiltered_items_count": 0,
            "unfiltered_batches_count": 0,
            "initial_unique_hashes_count": 0,
            "final_unique_hashes_count": 0,
        }

    @property
    def primary_key(self) -> Optional[TTableHintTemplate[TColumnNames]]:
        return self._primary_key

    @primary_key.setter
    def primary_key(self, value: str) -> None:
        # set key in incremental and data type transformers
        self._primary_key = value
        if self._transformers:
            for transform in self._transformers.values():
                transform.primary_key = value

    @classmethod
    def from_existing_state(
        cls, resource_name: str, cursor_path: str
    ) -> "Incremental[TCursorValue]":
        """Create Incremental instance from existing state."""
        state = Incremental._get_state(resource_name, cursor_path)
        i = cls(cursor_path, state["initial_value"])
        i.resource_name = resource_name
        return i

    def merge(self, other: "Incremental[TCursorValue]") -> "Incremental[TCursorValue]":
        """Create a new incremental instance which merges the two instances.
        Only properties which are not `None` from `other` override the current instance properties.

        This supports use cases with partial overrides, such as:
        >>> def my_resource(updated=incremental('updated', initial_value='1970-01-01'))
        >>>     ...
        >>>
        >>> my_resource(updated=incremental(initial_value='2023-01-01', end_value='2023-02-01'))
        """
        # func, resource name and primary key are not part of the dict
        kwargs = dict(
            self, last_value_func=self.last_value_func, primary_key=self._primary_key, lag=self.lag
        )
        for key, value in dict(
            other,
            last_value_func=other.last_value_func,
            primary_key=other.primary_key,
            lag=other.lag,
        ).items():
            if value is not None:
                kwargs[key] = value
        # preserve Generic param information
        if hasattr(self, "__orig_class__"):
            constructor = self.__orig_class__
        else:
            constructor = (
                other.__orig_class__ if hasattr(other, "__orig_class__") else other.__class__
            )
        constructor = extract_inner_type(constructor)
        merged = constructor(**kwargs)
        merged.resource_name = self.resource_name
        if other.resource_name:
            merged.resource_name = other.resource_name
        # also pass if resolved
        merged.__is_resolved__ = other.__is_resolved__
        merged.__exception__ = other.__exception__
        return merged  # type: ignore

    def copy(self) -> "Incremental[TCursorValue]":
        # merge creates a copy
        return self.merge(self)

    def get_cursor_column_name(self) -> Optional[str]:
        """Return the name of the cursor column if the cursor path resolves to a single column"""
        return extract_simple_field_name(self.cursor_path)

    def on_resolved(self) -> None:
        compile_path(self.cursor_path)
        if self.end_value is not None and self.initial_value is None:
            raise ConfigurationValueError(
                "Incremental `end_value` was specified without `initial_value`."
                "`initial_value` is required when using `end_value`."
            )
        # Ensure end value is "higher" than initial value
        try:
            if (
                self.end_value is not None
                and self.last_value_func([self.end_value, self.initial_value]) != self.end_value
            ):
                if self.last_value_func in (min, max):
                    adject = "higher" if self.last_value_func is max else "lower"
                    msg = (
                        f"Incremental `initial_value={self.initial_value}` is {adject} than"
                        f" `end_value={self.end_value}`. 'end_value' must be {adject} than"
                        " `initial_value`."
                    )
                else:
                    msg = (
                        f"Incremental `initial_value={self.initial_value}` is greater than"
                        f" `end_value={self.end_value}` as determined by the custom"
                        " `last_value_func`. The result of"
                        f" `{self.last_value_func.__name__}([end_value, initial_value])` must equal"
                        " `end_value`"
                    )
                raise ConfigurationValueError(msg)
        except ConfigurationValueError:
            raise
        except Exception as exc:
            raise ConfigurationValueError(
                f"Incremental `initial_value={self.initial_value}` and `end_value={self.end_value}`"
                " are not comparable. Make sure they are of the same type and tz-awareness: "
                + str(exc)
            ) from exc

    def parse_native_representation(self, native_value: Any) -> None:
        if isinstance(native_value, Incremental):
            if self is self.EMPTY:
                raise ValueError("Trying to resolve EMPTY Incremental")
            if native_value is self.EMPTY:
                raise ValueError(
                    "Do not use `EMPTY` Incremental as default or explicit values. "
                    "Pass `None` to reset an incremental."
                )
            merged = self.merge(native_value)
            self.cursor_path = merged.cursor_path
            self.initial_value = merged.initial_value
            self.last_value_func = merged.last_value_func
            self.end_value = merged.end_value
            self.resource_name = merged.resource_name
            self._primary_key = merged._primary_key
            self.allow_external_schedulers = merged.allow_external_schedulers
            self.row_order = merged.row_order
            self.lag = merged.lag
            # also copy the orig class to preserve cursor type
            if constructor := getattr(self, "__orig_class__", None):
                pass
            elif constructor := getattr(native_value, "__orig_class__", None):
                pass
            if constructor is not None:
                self.__orig_class__ = constructor
            self.__is_resolved__ = self.__is_resolved__
        else:  # TODO: Maybe check if callable(getattr(native_value, '__lt__', None))
            # Passing bare value `incremental=44` gets parsed as initial_value
            self.initial_value = native_value

    def get_state(self) -> IncrementalColumnState:
        """Returns or creates an Incremental state for a particular cursor column

        If end_value is set, a mock state is created that will be discarded after extract step
        Otherwise state is taken from current pipeline and will be persisted in it
        """
        if self.end_value is not None:
            # End value uses mock state. We don't want to write it.
            return {
                "initial_value": self.initial_value,
                "last_value": self.initial_value,
                "unique_hashes": [],
            }

        if not self.resource_name:
            raise IncrementalUnboundError(self.cursor_path)

        self._cached_state = Incremental._get_state(self.resource_name, self.cursor_path)
        if len(self._cached_state) == 0:
            # set the default like this, setdefault evaluates the default no matter if it is needed or not. and our default is heavy
            self._cached_state.update(
                {
                    "initial_value": self.initial_value,
                    "last_value": self.initial_value,
                    "unique_hashes": [],
                }
            )
        return self._cached_state

    @staticmethod
    def _get_state(resource_name: str, cursor_path: str) -> IncrementalColumnState:
        """Retrieve the sate from currently active pipeline"""
        state: IncrementalColumnState = (
            resource_state(resource_name).setdefault("incremental", {}).setdefault(cursor_path, {})
        )
        # if state params is empty
        return state

    @property
    def last_value(self) -> Optional[TCursorValue]:
        s = self.get_state()
        last_value: TCursorValue = s["last_value"]

        if self.lag:
            if self.last_value_func not in (max, min):
                logger.warning(
                    f"Lag on {self.resource_name} is only supported for max or min last_value_func."
                    f" Provided: {self.last_value_func}"
                )
            elif self.end_value is not None:
                logger.info(
                    f"Lag on {self.resource_name} is deactivated if end_value is set in"
                    " incremental."
                )
            elif last_value is not None:
                last_value = apply_lag(
                    self.lag, self.initial_value, last_value, self.last_value_func
                )

        return last_value

    def _transform_item(
        self, transformer: IncrementalTransform, row: TDataItem
    ) -> Optional[TDataItem]:
        row, self.start_out_of_range, self.end_out_of_range = transformer(row)
        # if we know that rows are ordered we can close the generator automatically
        # mind that closing pipe will not immediately close processing. it only closes the
        # generator so this page will be fully processed
        # TODO: we cannot close partially evaluated transformer gen. to implement that
        # we'd need to pass the source gen along with each yielded item and close this particular gen
        # NOTE: with that implemented we could implement add_limit as a regular transform having access to gen
        if self.can_close() and not self._bound_pipe.has_parent:
            self._bound_pipe.close()
        return row

    def get_incremental_value_type(self) -> Type[Any]:
        """Infers the type of incremental value from a class of an instance if those preserve the Generic arguments information."""
        return get_generic_type_argument_from_instance(self, self.initial_value)

    def _join_external_scheduler(self) -> None:
        """Detects existence of external scheduler from which `start_value` and `end_value` are taken. Detects Airflow and environment variables.
        The logical "start date" coming from external scheduler will set the `initial_value` in incremental. if additionally logical "end date" is
        present then also "end_value" will be set which means that resource state is not used and exactly this range of date will be loaded
        """
        # fit the pendulum into incremental type
        param_type = self.get_incremental_value_type()

        try:
            if param_type is not Any:
                data_type = py_type_to_sc_type(param_type)
        except Exception as ex:
            logger.warning(
                f"Specified Incremental last value type {param_type} is not supported. Please use"
                f" DateTime, Date, float, int or str to join external schedulers.({ex})"
            )
            return

        if param_type is Any:
            logger.warning(
                "Could not find the last value type of Incremental class participating in external"
                " schedule. Please add typing when declaring incremental argument in your resource"
                " or pass initial_value from which the type can be inferred."
            )
            return

        def _ensure_airflow_end_date(
            start_date: pendulum.DateTime, end_date: pendulum.DateTime
        ) -> Optional[pendulum.DateTime]:
            """if end_date is in the future or same as start date (manual run), set it to None so dlt state is used for incremental loading"""
            now = pendulum.now()
            if end_date is None or end_date > now or start_date == end_date:
                return now
            return end_date

        try:
            # we can move it to separate module when we have more of those
            from airflow.operators.python import get_current_context  # noqa

            context = get_current_context()
            start_date = context["data_interval_start"]
            end_date = _ensure_airflow_end_date(start_date, context["data_interval_end"])
            self.initial_value = coerce_from_date_types(data_type, start_date)
            if end_date is not None:
                self.end_value = coerce_from_date_types(data_type, end_date)
            else:
                self.end_value = None
            logger.info(
                f"Found Airflow scheduler: initial value: {self.initial_value} from"
                f" data_interval_start {context['data_interval_start']}, end value:"
                f" {self.end_value} from data_interval_end {context['data_interval_end']}"
            )
            return
        except TypeError as te:
            logger.warning(
                f"Could not coerce Airflow execution dates into the last value type {param_type}."
                f" ({te})"
            )
        except Exception:
            pass

        if start_value := os.environ.get("DLT_START_VALUE"):
            self.initial_value = coerce_value(data_type, "text", start_value)
            if end_value := os.environ.get("DLT_END_VALUE"):
                self.end_value = coerce_value(data_type, "text", end_value)
            else:
                self.end_value = None
            return

    def bind(self, pipe: SupportsPipe) -> "Incremental[TCursorValue]":
        """Called by pipe just before evaluation"""
        # bind the resource/pipe name
        if self.is_partial():
            raise IncrementalCursorPathMissing(pipe.name, None, None)
        self.resource_name = pipe.name
        self._bound_pipe = pipe
        # try to join external scheduler
        if self.allow_external_schedulers:
            self._join_external_scheduler()
        # set initial value from last value, in case of a new state those are equal
        self.start_value = self.last_value
        logger.info(
            f"Bind incremental on {self.resource_name} with initial_value: {self.initial_value},"
            f" start_value: {self.start_value}, end_value: {self.end_value}, func:"
            f" {self.last_value_func.__name__}, row_order: {self.row_order}, on_missing:"
            f" {self.on_cursor_value_missing}, range_start: {self.range_start}, range_end:"
            f" {self.range_end}"
        )
        # cache state
        self._cached_state = self.get_state()
        # Clear transforms so we get new instances
        self._transformers.clear()
        return self

    def can_close(self) -> bool:
        """Checks if incremental is out of range and can be closed.

        Returns true only when `row_order` was set and
        1. results are ordered ascending and are above upper bound (end_value)
        2. results are ordered descending and are below or equal lower bound (start_value)
        """
        # ordered ascending, check if we cross upper bound
        return (
            self.row_order == "asc"
            and self.end_out_of_range
            or self.row_order == "desc"
            and self.start_out_of_range
        )

    @classmethod
    def ensure_instance(cls, value: "TIncrementalConfig") -> "Incremental[TCursorValue]":
        if isinstance(value, Incremental):
            return value
        return cls(**value)

    def __str__(self) -> str:
        return (
            f"Incremental at 0x{id(self):x} for resource {self.resource_name} with cursor path:"
            f" {self.cursor_path} initial {self.initial_value} - {self.end_value} lv_func"
            f" {self.last_value_func}"
        )

    def _make_or_get_transformer(self, cls: Type[IncrementalTransform]) -> IncrementalTransform:
        if transformer := self._transformers.get(cls):
            return transformer
        transformer = self._transformers[cls] = cls(
            self.resource_name,
            self.cursor_path,
            self.initial_value,
            self.start_value,
            self.end_value,
            self.last_value_func,
            self._primary_key,
            set(self._cached_state["unique_hashes"]),
            self.on_cursor_value_missing,
            self.lag,
            self.range_start,
            self.range_end,
        )
        return transformer

    def _get_transform(self, items: TDataItems) -> IncrementalTransform:
        """Gets transform implementation that handles particular data item type"""
        # Assume list is all of the same type
        for item in items if isinstance(items, list) else [items]:
            if is_arrow_item(item):
                return self._make_or_get_transformer(ArrowIncremental)
            elif pandas is not None and isinstance(item, pandas.DataFrame):
                return self._make_or_get_transformer(ArrowIncremental)
            return self._make_or_get_transformer(JsonIncremental)
        return self._make_or_get_transformer(JsonIncremental)

    def __call__(self, rows: TDataItems, meta: Any = None) -> Optional[TDataItems]:
        # NOTE: we also forward empty lists, so special empty list types are preserved
        # example: MaterializedEmptyList
        if rows is None or (isinstance(rows, list) and len(rows) == 0):
            return rows

        # collect metrics
        self.custom_metrics["unfiltered_items_count"] += count_rows_in_items(rows)
        self.custom_metrics["unfiltered_batches_count"] += 1

        transformer = self._get_transform(rows)
        if isinstance(rows, list):
            rows = [
                item
                for item in (self._transform_item(transformer, row) for row in rows)
                if item is not None
            ]
            # return None if fully consumed like FilterItem (Incremental is just a very complicated FilterItem)
            if len(rows) == 0:
                rows = None
        else:
            rows = self._transform_item(transformer, rows)

        # ensure last_value maintains forward-only progression when lag is applied
        if self.lag and (cached_last_value := self._cached_state.get("last_value")):
            transformer.last_value = self.last_value_func(
                (transformer.last_value, cached_last_value)
            )
        # writing back state
        self._cached_state["last_value"] = transformer.last_value

        if transformer.boundary_deduplication:
            # compute hashes for new last rows
            # NOTE: object transform uses last_rows to pass rows to dedup, arrow computes
            #  hashes directly
            unique_hashes = set(
                transformer.compute_unique_value(row, self.primary_key)
                for row in transformer.last_rows
            )
            initial_hash_list = self._cached_state.get("unique_hashes")
            initial_hash_count = len(initial_hash_list) if initial_hash_list else 0
            self.custom_metrics["initial_unique_hashes_count"] = initial_hash_count

            # add directly computed hashes
            unique_hashes.update(transformer.unique_hashes)
            self._cached_state["unique_hashes"] = list(unique_hashes)
            final_hash_count = len(self._cached_state["unique_hashes"])
            self.custom_metrics["final_unique_hashes_count"] = final_hash_count

            self._check_duplicate_cursor_threshold(initial_hash_count, final_hash_count)
        else:
            self._cached_state["unique_hashes"] = []
        return rows

    def _check_duplicate_cursor_threshold(
        self, initial_hash_count: int, final_hash_count: int
    ) -> None:
        if initial_hash_count <= Incremental.duplicate_cursor_warning_threshold < final_hash_count:
            logger.warning(
                f"Large number of records ({final_hash_count}) sharing the same value of cursor"
                f" field '{self.cursor_path}' on resource '{self.resource_name}'. This can happen"
                " if the cursor field has a low resolution (e.g., only stores dates without"
                " times), causing many records to share the same cursor value. Consider using a"
                " cursor column with higher resolution to reduce the deduplication state size."
            )


Incremental.EMPTY = Incremental[Any]()
Incremental.EMPTY.__is_resolved__ = True

TIncrementalConfig = Union[Incremental[Any], IncrementalArgs]


class IncrementalResourceWrapper(ItemTransform[TDataItem, IncrementalCustomMetrics]):
    placement_affinity: ClassVar[float] = 1  # stick to end

    _incremental: Optional[Incremental[Any]] = None
    """Keeps the injectable incremental"""
    _from_hints: bool = False
    """If True, incremental was set explicitly from_hints"""
    _resource_name: str = None

    def __init__(self, primary_key: Optional[TTableHintTemplate[TColumnNames]] = None) -> None:
        """Creates a wrapper over a resource function that accepts Incremental instance in its argument to perform incremental loading.

        The wrapper delays instantiation of the Incremental to the moment of actual execution and is currently used by `dlt.resource` decorator.
        The wrapper explicitly (via `resource_name`) parameter binds the Incremental state to a resource state.
        Note that wrapper implements `FilterItem` transform interface and functions as a processing step in the before-mentioned resource pipe.

        Args:
            primary_key (TTableHintTemplate[TColumnKey], optional): A primary key to be passed to Incremental Instance at execution. Defaults to None.
        """
        self.primary_key = primary_key
        self.incremental_state: IncrementalColumnState = None
        self._allow_external_schedulers: bool = None
        self._bound_pipe: SupportsPipe = None

    @staticmethod
    def should_wrap(sig: inspect.Signature) -> bool:
        return IncrementalResourceWrapper.get_incremental_arg(sig) is not None

    @staticmethod
    def get_incremental_arg(sig: inspect.Signature) -> Optional[inspect.Parameter]:
        incremental_param: Optional[inspect.Parameter] = None
        for p in sig.parameters.values():
            annotation = extract_inner_type(
                resolve_single_annotation(p.annotation, globalns=globals())
            )
            if is_subclass(annotation, Incremental) or isinstance(p.default, Incremental):
                incremental_param = p
                break
        return incremental_param

    @staticmethod
    def inject_implicit_incremental_arg(
        incremental: Optional[Union[Incremental[Any], "IncrementalResourceWrapper"]],
        sig: inspect.Signature,
        func_args: Tuple[Any],
        func_kwargs: Dict[str, Any],
        fallback: Optional[Incremental[Any]] = None,
    ) -> Tuple[Tuple[Any], Dict[str, Any], Optional[Incremental[Any]]]:
        """Inject the incremental instance into function arguments
        if the function has an incremental argument without default in its signature and it is not already set in the arguments.

        Returns:
            Tuple of the new args, kwargs and the incremental instance that was injected (if any)
        """
        if isinstance(incremental, IncrementalResourceWrapper):
            incremental = incremental.incremental
        if not incremental:
            if not fallback:
                return func_args, func_kwargs, None
            incremental = fallback
        incremental_param = IncrementalResourceWrapper.get_incremental_arg(sig)
        if incremental_param:
            bound_args = sig.bind_partial(*func_args, **func_kwargs)
            if not bound_args.arguments.get(incremental_param.name):
                bound_args.arguments[incremental_param.name] = incremental
                return bound_args.args, bound_args.kwargs, incremental
        return func_args, func_kwargs, None

    def wrap(self, sig: inspect.Signature, func: TFun) -> TFun:
        """Wrap the callable to inject an `Incremental` object configured for the resource."""
        incremental_param = self.get_incremental_arg(sig)
        assert incremental_param, "Please use `should_wrap` to decide if to call this function"

        @wraps(func)
        def _wrap(*args: Any, **kwargs: Any) -> Any:
            p = incremental_param
            assert p is not None
            new_incremental: Incremental[Any] = None
            bound_args = sig.bind(*args, **kwargs)

            if p.name in bound_args.arguments:
                explicit_value = bound_args.arguments[p.name]
                if explicit_value is Incremental.EMPTY or p.default is Incremental.EMPTY:
                    raise ValueError(
                        "Do not use `EMPTY` Incremental as default or explicit values. "
                        "Pass `None` to reset an incremental."
                    )
                elif isinstance(explicit_value, Incremental):
                    # Explicit Incremental instance is merged with default
                    # allowing e.g. to only update initial_value/end_value but keeping default cursor_path
                    if isinstance(p.default, Incremental):
                        new_incremental = p.default.merge(explicit_value)
                    else:
                        new_incremental = explicit_value.copy()
                elif isinstance(p.default, Incremental):
                    # Passing only initial value explicitly updates the default instance
                    new_incremental = p.default.copy()
                    new_incremental.initial_value = explicit_value
            elif isinstance(p.default, Incremental):
                new_incremental = p.default.copy()

            if (not new_incremental or new_incremental.is_partial()) and not self._incremental:
                if is_optional_type(p.annotation):
                    bound_args.arguments[p.name] = None  # Remove partial spec
                    return func(*bound_args.args, **bound_args.kwargs)
                raise ValueError(
                    f"`{p.name}` incremental argument has no default. Please wrap its typing in"
                    " `Optional[]` to allow no incremental"
                )
            # pass Generic information from annotation to new_incremental
            if (
                new_incremental
                and not hasattr(new_incremental, "__orig_class__")
                and p.annotation
                and get_args(p.annotation)
            ):
                new_incremental.__orig_class__ = p.annotation

            # set the incremental only if not yet set or if it was passed explicitly
            # NOTE: the _incremental may be also set by applying hints to the resource see `set_template` in `DltResource`
            if (new_incremental and p.name in bound_args.arguments) or not self._incremental:
                self.set_incremental(new_incremental)
            if not self._incremental.is_resolved():
                self._incremental.resolve()
            # in case of transformers the bind will be called before this wrapper is set: because transformer is called for a first time late in the pipe
            if self._resource_name:
                # rebind internal _incremental from wrapper that already holds
                # instance of a Pipe
                self.bind(None)
            bound_args.arguments[p.name] = self._incremental
            return func(*bound_args.args, **bound_args.kwargs)

        return _wrap  # type: ignore

    @property
    def incremental(self) -> Optional[Incremental[Any]]:
        return self._incremental

    def set_incremental(
        self, incremental: Optional[TIncrementalConfig], from_hints: bool = False
    ) -> None:
        """Sets the incremental. If incremental was set from_hints, it can only be changed in the same manner"""
        if self._from_hints and not from_hints:
            # do not accept incremental if apply hints were used
            return
        if incremental is not None:
            incremental = Incremental.ensure_instance(incremental)
        self._from_hints = from_hints
        self._incremental = incremental

    @property
    def allow_external_schedulers(self) -> bool:
        """Allows the Incremental instance to get its initial and end values from external schedulers like Airflow"""
        if self._incremental:
            return self._incremental.allow_external_schedulers
        return self._allow_external_schedulers

    @allow_external_schedulers.setter
    def allow_external_schedulers(self, value: bool) -> None:
        self._allow_external_schedulers = value
        if self._incremental:
            self._incremental.allow_external_schedulers = value

    @property
    def custom_metrics(self) -> IncrementalCustomMetrics:
        """Returns custom metrics of the Incremental object itself if exists"""
        if self._incremental:
            return self._incremental.custom_metrics
        return {}

    def bind(self, pipe: SupportsPipe) -> "IncrementalResourceWrapper":
        # if pipe is None we are re-binding internal incremental
        pipe = pipe or self._bound_pipe
        self._bound_pipe = pipe
        self._resource_name = pipe.name
        if self._incremental:
            if self._allow_external_schedulers is not None:
                self._incremental.allow_external_schedulers = self._allow_external_schedulers
            self._incremental.bind(pipe)
        return self

    def __call__(self, item: TDataItems, meta: Any = None) -> Optional[TDataItems]:
        if not self._incremental:
            return item
        if self._incremental.primary_key is None:
            self._incremental.primary_key = self.primary_key
        elif self.primary_key is None:
            # propagate from incremental
            self.primary_key = self._incremental.primary_key
        return self._incremental(item, meta)


def incremental_config_to_instance(cfg: TIncrementalConfig) -> Incremental[Any]:
    if isinstance(cfg, Incremental):
        return cfg
    return Incremental(**cfg)


__all__ = [
    "Incremental",
    "IncrementalResourceWrapper",
    "IncrementalColumnState",
    "IncrementalCursorPathMissing",
    "IncrementalPrimaryKeyMissing",
    "IncrementalUnboundError",
    "TIncrementalconfig",
    "LastValueFunc",
    "TCursorValue",
]



from dlt._workspace.cli.ai.utils import (
    DEFAULT_AI_WORKBENCH_BRANCH,
    DEFAULT_AI_WORKBENCH_REPO,
)
from dlt._workspace.cli.ai.commands import (
    ai_status_command,
    ai_init_command,
    ai_mcp_run_command,
    ai_mcp_install_command,
    ai_secrets_list_command,
    ai_secrets_update_fragment_command,
    ai_secrets_view_redacted_command,
    ai_toolkit_info_command,
    ai_toolkit_install_command,
    ai_toolkit_list_command,
)

__all__ = [
    "DEFAULT_AI_WORKBENCH_BRANCH",
    "DEFAULT_AI_WORKBENCH_REPO",
    "ai_status_command",
    "ai_init_command",
    "ai_mcp_run_command",
    "ai_mcp_install_command",
    "ai_secrets_list_command",
    "ai_secrets_update_fragment_command",
    "ai_secrets_view_redacted_command",
    "ai_toolkit_info_command",
    "ai_toolkit_install_command",
    "ai_toolkit_list_command",
]



from functools import update_wrapper
import inspect
from typing import (
    AsyncIterable,
    cast,
    ClassVar,
    Callable,
    Iterable,
    Iterator,
    Generator,
    Type,
    Union,
    Any,
    Optional,
    Dict,
)

from dlt.common import logger
from dlt.common.configuration.inject import get_fun_spec, with_config
from dlt.common.configuration.resolve import inject_section
from dlt.common.configuration.specs import BaseConfiguration, known_sections
from dlt.common.configuration.specs.config_section_context import ConfigSectionContext
from dlt.common.reflection.inspect import isgeneratorfunction
from dlt.common.schema.utils import normalize_schema_name
from dlt.common.typing import (
    AnyFun,
    DictStrAny,
    StrAny,
    TDataItem,
    TDataItems,
    NoneType,
    Self,
    TypeVar,
)
from dlt.common.configuration.container import Container
from dlt.common.pipeline import (
    PipelineContext,
    StateInjectableContext,
    pipeline_state,
)
from dlt.common.utils import (
    flatten_list_or_items,
    get_callable_name,
    uniq_id,
    without_none,
    simple_repr,
)

from dlt.common.schema.typing import TTableSchema

from dlt.extract.utils import (
    make_schema_with_default_name,
    wrap_parallel_iterator,
    dynstr,
)
from dlt.extract.items import (
    DataItemWithMeta,
    TableNameMeta,
)
from dlt.extract.items_transform import (
    FilterItem,
    MapItem,
    YieldMapItem,
    ValidateItem,
    LimitItem,
    MetricsItem,
    ItemTransformFunc,
    ItemTransformFunctionWithMeta,
    MetricsFunctionWithMeta,
)
from dlt.extract.state import resource_state
from dlt.extract.pipe_iterator import ManagedPipeIterator
from dlt.extract.pipe import Pipe
from dlt.extract.hints import DltResourceHints, HintsMeta, TResourceHints
from dlt.extract.incremental import Incremental, IncrementalResourceWrapper
from dlt.extract.exceptions import (
    InvalidTransformerDataTypeGeneratorFunctionRequired,
    InvalidParentResourceDataType,
    InvalidParentResourceIsAFunction,
    InvalidResourceDataType,
    InvalidResourceDataTypeIsNone,
    InvalidTransformerGeneratorFunction,
    InvalidResourceDataTypeBasic,
    InvalidResourceDataTypeMultiplePipes,
    InvalidParallelResourceDataType,
    ParametrizedResourceUnbound,
    ResourceNameMissing,
    ResourceNotATransformer,
)
from dlt.extract.wrappers import wrap_additional_type


def with_table_name(item: TDataItems, table_name: str) -> DataItemWithMeta:
    """Marks `item` to be dispatched to table `table_name` when yielded from resource function."""
    return DataItemWithMeta(TableNameMeta(table_name), item)


def with_hints(
    item: TDataItems,
    hints: TResourceHints = None,
    create_table_variant: bool = False,
) -> DataItemWithMeta:
    """Marks `item` to update the resource with specified `hints`.

    Will create a separate variant of hints for a table if `name` is provided in `hints` and `create_table_variant` is set.

    Create `TResourceHints` with `make_hints`.
    Setting `table_name` will dispatch the `item` to a specified table, like `with_table_name`
    """
    return DataItemWithMeta(HintsMeta(hints, create_table_variant), item)


TDltResourceImpl = TypeVar("TDltResourceImpl", bound="DltResource", default="DltResource")


class DltResource(Iterable[TDataItem], DltResourceHints):
    """Implements dlt resource. Contains a data pipe that wraps a generating item and table schema that can be adjusted"""

    Empty: ClassVar["DltResource"] = None
    source_name: str
    """Name of the source that contains this instance of the source, set when added to DltResourcesDict"""
    section: str
    """A config section name"""
    SPEC: Type[BaseConfiguration]
    """A SPEC that defines signature of callable(parametrized) resource/transformer"""

    def __init__(
        self,
        pipe: Pipe,
        hints: TResourceHints,
        selected: bool,
        *,
        section: str = None,
        args_bound: bool = False,
    ) -> None:
        self.section = section
        self.selected = selected
        self._pipe = pipe
        self._args_bound = args_bound
        self._explicit_args: DictStrAny = None
        self.source_name = None
        self._parent: DltResource = None
        self._custom_metrics: Dict[str, Any] = {}
        super().__init__(hints)
        self._update_wrapper()

    @classmethod
    def from_data(
        cls,
        data: Any,
        name: str = None,
        section: str = None,
        hints: TResourceHints = None,
        selected: bool = True,
        data_from: "DltResource" = None,
        inject_config: bool = False,
    ) -> Self:
        """Creates an instance of DltResource from compatible `data` with a given `name` and `section`.

        Internally (in the most common case) a new instance of Pipe with `name` is created from `data` and
        optionally connected to an existing pipe `from_data` to form a transformer (dependent resource).

        If `inject_config` is set to True and data is a callable, the callable is wrapped in incremental and config
        injection wrappers.
        """
        if data is None:
            raise InvalidResourceDataTypeIsNone(name, data, NoneType)

        if isinstance(data, DltResource):
            r_: Self = data  # type: ignore[assignment]
        elif isinstance(data, Pipe):
            r_ = cls(data, hints, selected, section=section)
            if inject_config:
                r_._inject_config()
        else:
            if callable(data):
                name = name or get_callable_name(data)

            # if generator, take name from it
            if inspect.isgenerator(data):
                name = name or get_callable_name(data)  # type: ignore

            # name is mandatory
            if not name:
                raise ResourceNameMissing()

            # wrap additional types
            data = wrap_additional_type(data)

            # several iterable types are not allowed and must be excluded right away
            if isinstance(data, (str, dict)):
                raise InvalidResourceDataTypeBasic(name, data, type(data))

            # create resource from iterator, iterable or generator function
            if isinstance(data, (Iterable, Iterator, AsyncIterable)) or callable(data):
                # force unbounded transformer (data_From present) by setting empty pipe
                if data_from:
                    # verify transformer gen before populating steps
                    DltResource._ensure_valid_transformer_resource(name, data)
                    pipe = Pipe.from_data(name, data, parent=DltResource.Empty._pipe)
                else:
                    pipe = Pipe.from_data(name, data)

                r_ = cls(
                    pipe,
                    hints,
                    selected,
                    section=section,
                    args_bound=not callable(data),
                )
                if inject_config:
                    r_._inject_config()
            else:
                # some other data type that is not supported
                raise InvalidResourceDataType(
                    name,
                    data,
                    type(data),
                    f"The data type of supplied type is {type(data).__name__}",
                )
        if data_from:
            r_.pipe_data_from(data_from)
        return r_

    @property
    def custom_metrics(self) -> Dict[str, Any]:
        """Customizable resource metrics"""
        return self._custom_metrics

    @property
    def name(self) -> str:
        """Resource name inherited from the pipe"""
        return self._pipe.name

    def with_name(self, new_name: str, new_section: str = None) -> Self:
        """Clones the resource with a new name. Such resource keeps separate state and loads data to `new_name` table by default."""
        return self._clone(new_name=new_name, new_section=new_section, with_parent=True)

    @property
    def is_transformer(self) -> bool:
        """Checks if the resource is a transformer that takes data from another resource"""
        return self._pipe.has_parent

    @property
    def requires_args(self) -> bool:
        """Checks if resource has unbound arguments"""
        try:
            self._pipe.ensure_gen_bound()
            return False
        except (TypeError, ParametrizedResourceUnbound):
            return True

    @property
    def incremental(self) -> Optional[IncrementalResourceWrapper]:
        """Gets incremental transform if it is in the pipe"""
        return cast(
            Optional[IncrementalResourceWrapper],
            self._pipe.get_by_type(IncrementalResourceWrapper, Incremental),
        )

    @property
    def limit(self) -> Optional[LimitItem]:
        """Gets limit of pages/items on the resource"""
        return cast(Optional[LimitItem], self._pipe.get_by_type(LimitItem))

    @property
    def validator(self) -> Optional[ValidateItem]:
        """Gets validator transform if it is in the pipe"""
        return cast(Optional[ValidateItem], self._pipe.get_by_type(ValidateItem))

    @validator.setter
    def validator(self, validator: Optional[ValidateItem]) -> None:
        """Add/remove or replace the validator in pipe"""
        step_no = self._pipe.remove_by_type(ValidateItem)
        if validator:
            self.add_step(validator, insert_at=step_no if step_no >= 0 else None)

    @property
    def max_table_nesting(self) -> Optional[int]:
        """A schema hint for resource that sets the maximum depth of nested table above which the remaining nodes are loaded as structs or JSON."""
        max_nesting = self._hints.get("x-normalizer", {}).get("max_nesting")  # type: ignore[attr-defined]
        return max_nesting if isinstance(max_nesting, int) else None

    @max_table_nesting.setter
    def max_table_nesting(self, value: Optional[int]) -> None:
        normalizer = self._hints.setdefault("x-normalizer", {})  # type: ignore[typeddict-item]
        if value is None:
            normalizer.pop("max_nesting", None)
        else:
            normalizer["max_nesting"] = value

    def pipe_data_from(self, data_from: TDltResourceImpl) -> Self:
        """Replaces the parent execution pipe in the transformer by pipe in `data_from` which typically a resource.
        If resource was passed explicitly by the user, it will be stored as explicit parent
        """
        if self.is_transformer:
            DltResource._ensure_valid_transformer_resource(self.name, self._pipe.gen)
        else:
            raise ResourceNotATransformer(
                self.name, "Cannot pipe data into resource that is not a transformer."
            )
        parent_pipe = self._get_parent_pipe(self.name, data_from)
        self._pipe.parent = parent_pipe
        self._parent = data_from
        return self

    def add_pipe(self, data: Any) -> None:
        """Creates additional pipe for the resource from the specified data"""
        # TODO: (1) self resource cannot be a transformer (2) if data is resource both self must and it must be selected/unselected + cannot be transformer
        raise InvalidResourceDataTypeMultiplePipes(self.name, data, type(data))

    def select_tables(self, *table_names: Iterable[str]) -> Self:
        """For resources that dynamically dispatch data to several tables allows to select tables that will receive data, effectively filtering out other data items.

        Both `with_table_name` marker and data-based (function) table name hints are supported.
        """

        def _filter(item: TDataItem, meta: Any = None) -> bool:
            is_in_meta = isinstance(meta, TableNameMeta) and meta.table_name in table_names
            is_in_dyn = self._table_name_hint_fun and self._table_name_hint_fun(item) in table_names
            return is_in_meta or is_in_dyn

        # add filtering function at the end of pipe
        self.add_filter(_filter)
        return self

    def add_map(
        self, item_map: ItemTransformFunc[TDataItem], insert_at: int = None
    ) -> Self:  # noqa: A003
        """Adds mapping function defined in `item_map` to the resource pipe at position `inserted_at`

        `item_map` receives single data items, `dlt` will enumerate any lists of data items automatically

        Args:
            item_map (ItemTransformFunc[TDataItem]): A function taking a single data item and optional meta argument. Returns transformed data item.
            insert_at (int, optional): At which step in pipe to insert the mapping. Defaults to None which inserts after last step

        Returns:
            "DltResource": returns self
        """
        if insert_at is None:
            self._pipe.append_step(MapItem(item_map))
        else:
            self._pipe.insert_step(MapItem(item_map), insert_at)
        return self

    def add_metrics(
        self, metrics_f: MetricsFunctionWithMeta, insert_at: int = None
    ) -> Self:  # noqa: A003
        """Adds metrics collection function defined in `metrics_f` to the resource pipe at position `inserted_at`
        `metrics_f` receives data items, meta, and metrics dictionary to update in-place.
        Items are passed through unchanged.

        Args:
            metrics_f (MetricsFunctionWithMeta): A function taking items, meta argument, and metrics dict. Returns None.
            insert_at (int, optional): At which step in pipe to insert the metrics collector. Defaults to None which inserts after last step

        Returns:
            "DltResource": returns self
        """
        if insert_at is None:
            self._pipe.append_step(MetricsItem(metrics_f))
        else:
            self._pipe.insert_step(MetricsItem(metrics_f), insert_at)
        return self

    def add_yield_map(
        self,
        item_map: ItemTransformFunc[Iterator[TDataItem]],
        insert_at: int = None,
    ) -> Self:  # noqa: A003
        """Adds generating function defined in `item_map` to the resource pipe at position `inserted_at`

        `item_map` receives single data items, `dlt` will enumerate any lists of data items automatically. It may yield 0 or more data items and be used to
        ie. pivot an item into sequence of rows.

        Args:
            item_map (ItemTransformFunc[Iterator[TDataItem]]): A function taking a single data item and optional meta argument. Yields 0 or more data items.
            insert_at (int, optional): At which step in pipe to insert the generator. Defaults to None which inserts after last step

        Returns:
            "DltResource": returns self
        """
        if insert_at is None:
            self._pipe.append_step(YieldMapItem(item_map))
        else:
            self._pipe.insert_step(YieldMapItem(item_map), insert_at)
        return self

    def add_filter(
        self, item_filter: ItemTransformFunc[bool], insert_at: int = None
    ) -> Self:  # noqa: A003
        """Adds filter defined in `item_filter` to the resource pipe at position `inserted_at`

        `item_filter` receives single data items, `dlt` will enumerate any lists of data items automatically

        Args:
            item_filter (ItemTransformFunc[bool]): A function taking a single data item and optional meta argument. Returns bool. If True, item is kept
            insert_at (int, optional): At which step in pipe to insert the filter. Defaults to None which inserts after last step
        Returns:
            "DltResource": returns self
        """
        if insert_at is None:
            self._pipe.append_step(FilterItem(item_filter))
        else:
            self._pipe.insert_step(FilterItem(item_filter), insert_at)
        return self

    def add_limit(
        self,
        max_items: Optional[int] = None,
        max_time: Optional[float] = None,
        count_rows: Optional[bool] = False,
    ) -> Self:  # noqa: A003
        """Limit the number of items that will be processed by the resource: by count and by time. By default, dlt counts
        number of "yields/batches/pages" not the number of rows inside.

        For incremental resources that return rows in deterministic order, you can use this function to process large load
        in batches.

         Notes:
             1. Transformers won't be limited. They should process all the data they receive fully to avoid inconsistencies in generated datasets.
             2. Each yielded item may contain several rows. By default `add_limit` only limits the "number of yields", not the total number of rows.
             3. Empty pages/yields are also counted. Use `count_rows` to skip empty pages.

        Args:
             max_items (int): The maximum number of items (not rows!) to yield, set to None for no limit
             max_time (float): The maximum number of seconds for this generator to run after it was opened, set to None for no limit
             count_rows (bool): Default: false.
                Count rows instead of pages. Note that if resource yields pages of rows, last page will not be trimmed and more rows that expected will be received.
         Returns:
             "DltResource": returns self
        """

        if self.is_transformer:
            logger.warning(
                f"Setting add_limit to a transformer {self.name} has no effect. Set the limit on"
                " the top level resource."
            )
        else:
            # remove existing limit if any
            self._pipe.remove_by_type(LimitItem)
            self.add_step(LimitItem(max_items=max_items, max_time=max_time, count_rows=count_rows))

        return self

    def parallelize(self) -> Self:
        """Wraps the resource to execute each item in a threadpool to allow multiple resources to extract in parallel.

        The resource must be a generator or generator function or a transformer function.
        """
        if (
            not inspect.isgenerator(self._pipe.gen)
            and not (callable(self._pipe.gen) and isgeneratorfunction(self._pipe.gen))
            and not (callable(self._pipe.gen) and self.is_transformer)
        ):
            raise InvalidParallelResourceDataType(self.name, self._pipe.gen, type(self._pipe.gen))

        self._pipe.replace_gen(wrap_parallel_iterator(self._pipe.gen))  # type: ignore  # TODO
        return self

    def add_step(
        self,
        item_transform: ItemTransformFunctionWithMeta[TDataItems],
        insert_at: int = None,
    ) -> Self:  # noqa: A003
        if insert_at is None:
            self._pipe.append_step(item_transform)
        else:
            self._pipe.insert_step(item_transform, insert_at)
        return self

    def _remove_incremental_step(self) -> None:
        self._pipe.remove_by_type(Incremental, IncrementalResourceWrapper)

    def set_incremental(
        self,
        new_incremental: Union[Incremental[Any], IncrementalResourceWrapper],
        from_hints: bool = False,
    ) -> Optional[Union[Incremental[Any], IncrementalResourceWrapper]]:
        """Set/replace the incremental transform for the resource.

        Args:
            new_incremental: The Incremental instance/hint to set or replace
            from_hints: If the incremental is set from hints. Defaults to False.
        """
        if new_incremental is Incremental.EMPTY:
            new_incremental = None
        incremental = self.incremental
        if incremental is not None:
            # if isinstance(new_incremental, Mapping):
            #     new_incremental = Incremental.ensure_instance(new_incremental)

            if isinstance(new_incremental, IncrementalResourceWrapper):
                # Completely replace the wrapper
                self._remove_incremental_step()
                self.add_step(new_incremental)
            elif isinstance(incremental, IncrementalResourceWrapper):
                incremental.set_incremental(new_incremental, from_hints=from_hints)
            else:
                self._remove_incremental_step()
                # re-add the step
                incremental = None
        if incremental is None:
            # if there's no wrapper add incremental as a transform
            if new_incremental:
                if not isinstance(new_incremental, IncrementalResourceWrapper):
                    new_incremental = Incremental.ensure_instance(new_incremental)
                self.add_step(new_incremental)
        return new_incremental

    def _set_hints(
        self, table_schema_template: TResourceHints, create_table_variant: bool = False
    ) -> None:
        super()._set_hints(table_schema_template, create_table_variant)
        # validators and incremental apply only to resource hints
        if not create_table_variant:
            # try to late assign incremental
            if table_schema_template.get("incremental") is not None:
                incremental = self.set_incremental(
                    table_schema_template["incremental"], from_hints=True
                )
            else:
                incremental = self.incremental

            if incremental:
                primary_key = table_schema_template.get("primary_key", incremental.primary_key)
                if primary_key is not None:
                    incremental.primary_key = primary_key

            if table_schema_template.get("validator") is not None:
                self.validator = table_schema_template["validator"]

    def compute_table_schema(self, item: TDataItem = None, meta: Any = None) -> TTableSchema:
        incremental: Optional[Union[Incremental[Any], IncrementalResourceWrapper]] = (
            self.incremental
        )
        if incremental and "incremental" not in self._hints:
            if isinstance(incremental, IncrementalResourceWrapper):
                incremental = incremental.incremental
                if incremental:
                    self._hints["incremental"] = incremental

        table_schema = super().compute_table_schema(item, meta)

        return table_schema

    def bind(self, *args: Any, **kwargs: Any) -> Self:
        """Binds the parametrized resource to passed arguments. Modifies resource pipe in place. Does not evaluate generators or iterators."""
        if self._args_bound:
            raise TypeError(
                f"Parametrized resource `{self.name}` is not callable. You can call and pass"
                " arguments to a parametrized resource only once. Make sure you didn't call this"
                " resource before."
            )

        orig_gen = self._pipe.gen
        gen = self._pipe.bind_gen(*args, **kwargs)
        if isinstance(gen, DltResource):
            # the resource returned resource: update in place
            # TODO: find a better way to modify resource when called
            old_pipe = self._pipe
            self.__dict__.clear()
            self.__dict__.update(gen.__dict__)
            # keep old pipe instance
            self._pipe = old_pipe
            self._pipe.__dict__.clear()
            # write props from new pipe instance
            self._pipe.__dict__.update(gen._pipe.__dict__)
        # elif isinstance(gen, Pipe):
        #     # the resource returned pipe: just replace pipe
        #     self._pipe.__dict__.clear()
        #     # write props from new pipe instance
        #     self._pipe.__dict__.update(gen.__dict__)
        else:
            self._args_bound = True
        self._set_explicit_args(orig_gen, None, *args, **kwargs)  # type: ignore
        return self

    @property
    def args_bound(self) -> bool:
        """Returns true if resource the parameters are bound to values. Such resource cannot be further called.
        Note that resources are lazily evaluated and arguments are only formally checked. Configuration
        was not yet injected as well.
        """
        return self._args_bound

    @property
    def explicit_args(self) -> StrAny:
        """Returns a dictionary of arguments used to parametrize the resource. Does not include defaults and injected args."""
        if not self._args_bound:
            raise TypeError(f"Resource `{self.name}` is not yet parametrized")
        return self._explicit_args

    @property
    def state(self) -> StrAny:
        """Gets resource-scoped state from the active pipeline. PipelineStateNotAvailable is raised if pipeline context is not available"""
        with inject_section(self._get_config_section_context()):
            return resource_state(self.name)

    def __call__(self: TDltResourceImpl, *args: Any, **kwargs: Any) -> TDltResourceImpl:
        """Binds the parametrized resources to passed arguments. Creates and returns a bound resource. Generators and iterators are not evaluated."""
        if self._args_bound:
            raise TypeError(
                f"Parametrized resource `{self.name}` is not callable. You can call and pass"
                " arguments to a parametrized resource only once. Make sure you didn't call this"
                " resource before."
            )
        # detect if name should be renamed
        bound_args = None
        if isinstance(self.name, dynstr) or isinstance(self.section, dynstr):
            _, _, bound_args = self._pipe.sim_gen(*args, **kwargs)
        new_name = self.name(bound_args.arguments) if isinstance(self.name, dynstr) else None
        new_section = (
            self.section(bound_args.arguments) if isinstance(self.section, dynstr) else None
        )

        r = self._clone(new_name=new_name, new_section=new_section)
        return r.bind(*args, **kwargs)

    def __or__(self, transform: Union["DltResource", AnyFun]) -> "DltResource":
        """Allows to pipe data from across resources and transform functions with | operator
        This is the LEFT side OR so the self may be resource or transformer
        """
        # print(f"{resource.name} | {self.name} -> {resource.name}[{resource.is_transformer}]")
        if isinstance(transform, DltResource):
            transform.pipe_data_from(self)
            # return transformed resource for chaining
            return transform
        else:
            # map or yield map
            if isgeneratorfunction(transform):
                return self.add_yield_map(transform)
            else:
                return self.add_map(transform)

    def __ror__(self, data: Union[Iterable[Any], Iterator[Any]]) -> Self:
        """Allows to pipe data from across resources and transform functions with | operator
        This is the RIGHT side OR so the self may not be a resource and the LEFT must be an object
        that does not implement | ie. a list
        """
        self.pipe_data_from(self.from_data(data, name="iter_" + uniq_id(4)))
        return self

    def __iter__(self) -> Generator[TDataItem, None, None]:
        """Opens iterator that yields the data items from the resources in the same order as in Pipeline class.

        A read-only state is provided, initialized from active pipeline state. The state is discarded after the iterator is closed.
        """
        # use the same state dict when opening iterator and when iterator is iterated
        container = Container()
        state, _ = pipeline_state(container, {})
        state_context = StateInjectableContext(state=state)
        section_context = self._get_config_section_context()

        # create single resource source
        from dlt.common.schema import Schema
        from dlt.extract.source import (
            _DltSingleSource,
            SourceInjectableContext,
            SourceSchemaInjectableContext,
        )

        source = _DltSingleSource(Schema(section_context.source_state_key), self.section, [self])
        source_context = SourceInjectableContext(source)
        schema_context = SourceSchemaInjectableContext(source.discover_schema())

        # managed pipe iterator will set the context on each call to  __next__
        with (
            inject_section(section_context),
            Container().injectable_context(state_context),
            Container().injectable_context(source_context),
            Container().injectable_context(schema_context),
        ):
            pipe_iterator: ManagedPipeIterator = ManagedPipeIterator.from_pipes([self._pipe])  # type: ignore

        pipe_iterator.set_context([state_context, section_context, schema_context, source_context])
        _iter = map(lambda item: item.item, pipe_iterator)
        return flatten_list_or_items(_iter)

    def _set_explicit_args(
        self, f: AnyFun, sig: inspect.Signature = None, *args: Any, **kwargs: Any
    ) -> None:
        try:
            sig = sig or inspect.signature(f)
            self._explicit_args = sig.bind_partial(*args, **kwargs).arguments
        except Exception:
            pass

    def _eject_config(self) -> bool:
        """Unwraps the pipe generator step from config injection and incremental wrappers by restoring the original step.

        Removes the step with incremental wrapper. Should be used before a subsequent _inject_config is called on the
        same pipe to successfully wrap it with new incremental and config injection.
        Note that resources with bound arguments cannot be ejected.

        """
        if not self._pipe.is_empty and not self._args_bound:
            orig_gen = getattr(self._pipe.gen, "__GEN__", None)
            if orig_gen:
                self._remove_incremental_step()
                self._pipe.replace_gen(orig_gen)
                return True
        return False

    def _inject_config(self, incremental_from_hints_override: Optional[bool] = None) -> Self:
        """Wraps the pipe generation step in incremental and config injection wrappers and adds pipe step with
        Incremental transform.
        """
        gen = self._pipe.gen
        if not callable(gen):
            return self

        incremental: IncrementalResourceWrapper = None
        sig = inspect.signature(gen)
        if IncrementalResourceWrapper.should_wrap(sig):
            incremental = IncrementalResourceWrapper(self._hints.get("primary_key"))
            if incr_hint := self._hints.get("incremental"):
                incremental.set_incremental(
                    incr_hint,
                    from_hints=(
                        incremental_from_hints_override
                        if incremental_from_hints_override is not None
                        else True
                    ),
                )
            incr_f = incremental.wrap(sig, gen)
            self.set_incremental(incremental)
        else:
            incr_f = gen
        resource_sections = (known_sections.SOURCES, self.section, self.name)
        # function should have associated SPEC
        spec = get_fun_spec(gen)
        # standalone resource will prefer existing section context when resolving config values
        # this lets the source to override those values and provide common section for all config values for resources present in that source
        # for autogenerated spec do not include defaults
        conf_f = with_config(
            incr_f,
            spec=spec,
            sections=resource_sections,
            sections_merge_style=ConfigSectionContext.resource_merge_style,
        )
        if conf_f != gen:
            self._pipe.replace_gen(conf_f)
            # storage the original generator to be able to eject config and incremental wrapper
            # when resource is cloned
            setattr(conf_f, "__GEN__", gen)  # noqa: B010
        return self

    def _clone(
        self, *, new_name: str = None, new_section: str = None, with_parent: bool = False
    ) -> Self:
        """Creates a deep copy of a current resource, optionally renaming the resource.
        The clone will not be part of the source and will not include the custom resource metrics.
        """
        pipe = self._pipe
        if pipe and not pipe.is_empty:
            pipe = pipe._clone(new_name=new_name)
            # assign new instance id
            pipe.instance_id = uniq_id()

        # incremental and parent are already in the pipe (if any)
        incremental = self.incremental
        if isinstance(incremental, IncrementalResourceWrapper):
            incremental_from_hints: Optional[bool] = incremental._from_hints
        else:
            incremental_from_hints = None
        cloned_r = self.__class__(
            pipe,
            self._clone_hints(self._hints),
            selected=self.selected,
            section=new_section or self.section,
            args_bound=self._args_bound,
        )

        # keep the parent
        cloned_r._parent = self._parent
        # also clone parent if requested
        if with_parent and cloned_r._parent and not cloned_r._parent._pipe.is_empty:
            parent_new_name = new_name
            # if we are renaming the resource, then also rename the parent
            if new_name:
                # if original names contained common element, then continue that pattern
                if self.name in self._parent.name:
                    parent_new_name = cloned_r._parent.name.replace(self.name, new_name)
                else:
                    parent_new_name = f"{self._parent.name}_{new_name}"

            cloned_r._parent = cloned_r._parent._clone(
                new_name=parent_new_name, new_section=new_section, with_parent=with_parent
            )
            cloned_r._pipe.parent = cloned_r._parent._pipe

        # try to eject and then inject configuration and incremental wrapper when resource is cloned
        # this makes sure that a take config values from a right section and wrapper has a separated
        # instance in the pipeline
        if cloned_r._eject_config():
            cloned_r._inject_config(incremental_from_hints_override=incremental_from_hints)
        return cloned_r

    def _update_wrapper(self) -> None:
        """Update wrapper from callable pipe gen to behave as functools wrapper"""
        if self._pipe.is_empty:
            return
        gen = self._pipe.gen
        if not callable(gen):
            return
        update_wrapper(self, gen)
        # also change the signature return annotation to dlt source class
        sig = inspect.signature(gen).replace(return_annotation=self.__class__)
        self.__signature__ = sig

    def _get_config_section_context(self) -> ConfigSectionContext:
        container = Container()
        proxy = container[PipelineContext]
        pipeline = None if not proxy.is_active() else proxy.pipeline()
        if pipeline:
            pipeline_name = pipeline.pipeline_name
        else:
            pipeline_name = None
        if pipeline:
            default_schema_name = pipeline.default_schema_name
        else:
            default_schema_name = None
        if not default_schema_name and pipeline_name:
            default_schema_name = make_schema_with_default_name(pipeline.pipeline_name)
        return ConfigSectionContext(
            pipeline_name=pipeline_name,
            # do not emit middle config section to not overwrite the resource section
            # only sources emit middle config section
            sections=(
                known_sections.SOURCES,
                self.section or pipeline_name or "s_" + uniq_id(4),
                self.name,
            ),
            source_state_key=(
                self.source_name
                or default_schema_name
                or (normalize_schema_name(self.section) if self.section else None)
                or "schema_" + uniq_id(4)
            ),
        )

    def __repr__(self) -> str:
        # TODO add a mechanism to truncate the repr of some hints
        # TODO expand information about steps
        # TODO ensure consistent ordering between object __repr__
        # TODO verify that attributes are inexpensive to compute
        # TODO add a toggle for kwargs that are not valid kwargs, but
        # helpful for debugging

        limit = None
        if (step_id := self._pipe.find(LimitItem)) != -1:
            step: LimitItem = self._pipe.steps[step_id]  # type: ignore[assignment]
            # step limit in chunks
            limit = step.limit(1)

        hints = self._hints if self._hints is not None else {}
        kwargs = {
            "name": self.name,
            "section": self.section,
            "table_name": hints.get("table_name"),
            "primary_key": hints.get("primary_key"),
            "merge_key": hints.get("merge_key"),
            "columns": "{...}" if hints.get("columns") else None,
            "parent_table_name": hints.get("parent_table_name"),
            "references": "{...}" if hints.get("references") else None,
            "nested_hints": "{...}" if hints.get("nested_hints") else None,
            "limit": limit,  # NOTE not a valid kwarg for `@dlt.resource`
            "max_table_nesting": hints.get("max_table_nesting"),
            "write_disposition": hints.get("write_disposition"),
            "table_format": hints.get("table_format"),
            "file_format": hints.get("file_format"),
            "schema_contract": "{...}" if hints.get("schema_contract") else None,
            "incremental": self.incremental,
            "validator": self.validator,
        }
        if len(self._pipe.steps) > 1:
            # NOTE both are not valid kwargs for `@dlt.resource`
            kwargs["n_steps"] = len(self._pipe.steps)
            kwargs["steps"] = [type(step).__name__ for step in self._pipe.steps]
        # the name isn't `DltResource` because it's not the main entrypoint
        # to create a resource
        if self.is_transformer:
            kwargs["parent"] = self._parent.name
            return simple_repr("@dlt.transformer", **without_none(kwargs))
        else:
            return simple_repr("@dlt.resource", **without_none(kwargs))

    def __str__(self) -> str:
        info = f"DltResource `{self.name}`"
        if self.section:
            info += f" in section `{self.section}`"
        if self.source_name:
            info += f" added to source `{self.source_name}`:"
        else:
            info += ":"

        if self.is_transformer:
            info += (
                "\nThis resource is a transformer and takes data items from"
                f" `{self._pipe.parent.name}`"
            )
        else:
            if self._pipe.is_data_bound:
                if self.requires_args:
                    head_sig = inspect.signature(self._pipe.gen)  # type: ignore
                    info += (
                        "\nThis resource is parametrized and takes the following arguments"
                        f" `{head_sig}`. You must call this resource before loading."
                    )
                else:
                    info += (
                        "\nIf you want to see the data items in the resource you must iterate it or"
                        " convert to list ie. `list(resource)`. Note that, like any iterator, you"
                        " can iterate the resource only once."
                    )
            else:
                info += "\nThis resource is not bound to the data"
        info += f"\nInstance: info: (data pipe id:{id(self._pipe)}) at {id(self)}"
        return info

    @staticmethod
    def _ensure_valid_transformer_resource(name: str, data: Any) -> None:
        # resource must be a callable with single argument
        if callable(data):
            valid_code = DltResource.validate_transformer_generator_function(data)
            if valid_code != 0:
                raise InvalidTransformerGeneratorFunction(
                    name, get_callable_name(data), inspect.signature(data), valid_code
                )
        else:
            raise InvalidTransformerDataTypeGeneratorFunctionRequired(name, data, type(data))

    @staticmethod
    def _get_parent_pipe(name: str, data_from: Union["DltResource", Pipe]) -> Pipe:
        if isinstance(data_from, DltResource):
            return data_from._pipe
        else:
            # if this is generator function provide nicer exception
            if callable(data_from):
                raise InvalidParentResourceIsAFunction(name, get_callable_name(data_from))
            else:
                raise InvalidParentResourceDataType(name, data_from, type(data_from))

    @staticmethod
    def validate_transformer_generator_function(f: AnyFun) -> int:
        sig = inspect.signature(f)
        if len(sig.parameters) == 0:
            return 1
        # transformer may take only one positional only argument
        pos_only_len = sum(1 for p in sig.parameters.values() if p.kind == p.POSITIONAL_ONLY)
        if pos_only_len > 1:
            return 2
        first_ar = next(iter(sig.parameters.values()))
        # and pos only must be first
        if pos_only_len == 1 and first_ar.kind != first_ar.POSITIONAL_ONLY:
            return 2
        # first arg must be positional or kw_pos
        if first_ar.kind not in (first_ar.POSITIONAL_ONLY, first_ar.POSITIONAL_OR_KEYWORD):
            return 3
        return 0


# produce Empty resource singleton
DltResource.Empty = DltResource(Pipe(None), None, False)
TUnboundDltResource = Callable[..., DltResource]



from typing import List, Optional, Type, Literal, Union
from types import ModuleType

from dlt.common.typing import StrAny, TypedDict
from dlt.common.normalizers.naming import NamingConvention

TNamingConventionReferenceArg = Union[str, Type[NamingConvention], ModuleType]


TRowIdType = Literal["random", "row_hash", "key_hash"]


class TJSONNormalizer(TypedDict, total=False):
    module: str
    config: Optional[StrAny]  # config is a free form and is validated by `module`


class TNormalizersConfig(TypedDict, total=False):
    names: str
    allow_identifier_change_on_table_with_data: Optional[bool]
    use_break_path_on_normalize: Optional[bool]
    """Post 1.4.0 to allow table and column names that contain table separators"""
    detections: Optional[List[str]]
    json: TJSONNormalizer



"""Source that loads tables form any SQLAlchemy supported database, supports batching requests and incremental loads."""

from typing import Callable, Dict, List, Optional, Type, Union, Iterable, Any

import dlt
from dlt.common.configuration.specs import ConnectionStringCredentials
from dlt.common.schema.typing import TWriteDispositionConfig
from dlt.common.libs.sql_alchemy import MetaData, Table, Engine

from dlt.common.typing import TColumnNames
from dlt.extract import DltResource, Incremental, decorators

from .helpers import (
    _execute_table_adapter,
    default_engine_adapter_callback,
    table_rows,
    engine_from_credentials,
    remove_nullability_adapter,
    BaseTableLoader,
    TableLoader,
    ConnectorXTableLoader,
    register_table_loader_backend,
    get_table_loader_class,
    TableBackend,
    SqlTableResourceConfiguration,
    _detect_precision_hints_deprecated,
    TQueryAdapter,
    TTableAdapter,
)
from .schema_types import (
    table_to_resource_hints,
    ReflectionLevel,
    TTypeAdapter,
)


@decorators.source
def sql_database(
    credentials: Union[ConnectionStringCredentials, Engine, str] = dlt.secrets.value,
    schema: Optional[str] = dlt.config.value,
    metadata: Optional[MetaData] = None,
    table_names: Optional[List[str]] = dlt.config.value,
    chunk_size: int = 50000,
    backend: TableBackend = "sqlalchemy",
    detect_precision_hints: Optional[bool] = False,
    reflection_level: Optional[ReflectionLevel] = "full",
    defer_table_reflect: Optional[bool] = None,
    table_adapter_callback: Optional[TTableAdapter] = None,
    backend_kwargs: Dict[str, Any] = None,
    include_views: bool = False,
    type_adapter_callback: Optional[TTypeAdapter] = None,
    query_adapter_callback: Optional[TQueryAdapter] = None,
    resolve_foreign_keys: bool = False,
    engine_adapter_callback: Optional[Callable[[Engine], Engine]] = None,
    engine_kwargs: Optional[Dict[str, Any]] = None,
    table_loader_class: Optional[Type[BaseTableLoader]] = None,
) -> Iterable[DltResource]:
    """
    A dlt source which loads data from an SQL database using SQLAlchemy.
    Resources are automatically created for each table in the schema or from the given list of tables.

    Args:
        credentials (Union[ConnectionStringCredentials, Engine, str]): Database credentials or an `sqlalchemy.Engine` instance.

        schema (Optional[str]): Name of the database schema to load (if different from default).

        metadata (Optional[MetaData]): Optional `sqlalchemy.MetaData` instance. `schema` argument is ignored when this is used.

        table_names (Optional[List[str]]): A list of table names to load. By default, all tables in the schema are loaded.

        chunk_size (int): Number of rows yielded in one batch. SQL Alchemy will create additional internal rows buffer twice the chunk size.

        backend (TableBackend): Type of backend to generate table data. One of: "sqlalchemy", "pyarrow", "pandas" and "connectorx".
            "sqlalchemy" yields batches as lists of Python dictionaries, "pyarrow" and "connectorx" yield batches as arrow tables, "pandas" yields pandas DataFrames.
            "sqlalchemy" is the default and does not require additional dependencies, "pyarrow" creates stable destination schemas with correct data types,
            "connectorx" is typically the fastest but ignores the "chunk_size" so you must deal with large tables yourself.

        detect_precision_hints (Optional[bool]): Deprecated. Use `reflection_level`. Set column precision and scale hints for supported data types in the target schema based on the columns in the source tables.
            This is disabled by default.

        reflection_level (Optional[ReflectionLevel]): Specifies how much information should be reflected from the source database schema.
            "minimal": Only table names, nullability and primary keys are reflected. Data types are inferred from the data.
            "full" (default): Data types will be reflected on top of "minimal". `dlt` will coerce the data into reflected types if necessary.
            "full_with_precision": Sets precision and scale on supported data types (ie. decimal, text, binary). Creates big and regular integer types.

        defer_table_reflect (Optional[bool]): Will connect and reflect table schema only when yielding data. Requires `table_names` to be explicitly passed.
            Enable this option when running on Airflow and other orchestrators that create execution DAGs. When True, schema is decided during execution,
            which may override `query_adapter_callback` modifications or `apply_hints`.

        table_adapter_callback (Optional[TTableAdapter]): Receives each reflected table. May be used to modify the list of columns that will be selected.

        backend_kwargs (Dict[str, Any]): kwargs passed to table backend ie. "conn" is used to pass specialized connection string to connectorx.

        include_views (bool): Reflect views as well as tables. Note view names included in `table_names` are always included regardless of this setting.

        type_adapter_callback (Optional[TTypeAdapter]): Callable to override type inference when reflecting columns.
            Argument is a single sqlalchemy data type (`TypeEngine` instance) and it should return another sqlalchemy data type, or `None` (type will be inferred from data)

        query_adapter_callback (Optional[TQueryAdapter]): Callable to override the SELECT query used to fetch data from the table.
            The callback receives the sqlalchemy `Select` and corresponding `Table`, 'Incremental` and `Engine` objects and should return the modified `Select` or `Text`.

        resolve_foreign_keys (bool): Translate foreign keys in the same schema to `references` table hints.
            May incur additional database calls as all referenced tables are reflected.

        engine_adapter_callback (Optional[Callable[[Engine], Engine]]): Callback to configure, modify an Engine instance that will be used to open a connection ie. to
            set transaction isolation level.

        engine_kwargs (Optional[Dict[str, Any]]): Optional SQLAlchemy engine keyword arguments passed directly to `sqlalchemy.create_engine()`.
            They always affect table reflection and also data loading if SQLAlchemy backend is used (default). If other backend is used, pass equivalent idiomatic params to backend_kwargs.

        table_loader_class (Optional[Type[BaseTableLoader]]): Custom table loader class to use for loading data.
            Subclass `TableLoader` to customize row loading within the SQLAlchemy ecosystem (e.g. pagination,
            retry logic), or subclass `BaseTableLoader` for entirely different backends. When not provided,
            the default loader is selected based on the `backend` parameter.

    Yields:
        DltResource: DLT resources for each table to be loaded.
    """
    # detect precision hints is deprecated
    _detect_precision_hints_deprecated(detect_precision_hints)

    if detect_precision_hints:
        reflection_level = "full_with_precision"
    else:
        reflection_level = reflection_level or "minimal"

    engine_kwargs = engine_kwargs or {}
    engine = engine_from_credentials(
        credentials,
        may_dispose_after_use=False,
        **engine_kwargs,
    )
    engine.execution_options(stream_results=True, max_row_buffer=2 * chunk_size)
    if engine_adapter_callback:
        engine = engine_adapter_callback(engine)
    metadata = metadata or MetaData(schema=schema)
    default_engine_adapter_callback(engine, metadata)

    if defer_table_reflect:
        if not table_names:
            raise ValueError("You must pass `table_names` to defer table reflection")
        table_infos = [(schema, table) for table in table_names]
    else:
        # reflect tables
        metadata.reflect(
            bind=engine,
            views=include_views or bool(table_names),  # Specified view names are always reflected
            only=table_names if table_names else None,
            resolve_fks=resolve_foreign_keys,
        )
        tables = list(metadata.tables.values())
        # Some extra tables may be reflected in metadata due to foreign keys
        table_infos = [
            (table.schema, table.name)
            for table in tables
            if table_names is None or table.name in table_names
        ]

    # dispose the reflection engine — MetaData no longer needs it and each
    # sql_table() resource creates its own engine from credentials.
    # skipped for externally-provided Engine instances (user manages lifecycle).
    if not isinstance(credentials, Engine):
        engine.dispose()

    for table_schema, table_name in table_infos:
        yield sql_table(
            credentials=credentials,
            table=table_name,
            schema=table_schema,
            metadata=metadata,
            chunk_size=chunk_size,
            backend=backend,
            reflection_level=reflection_level,
            defer_table_reflect=defer_table_reflect,
            table_adapter_callback=table_adapter_callback,
            backend_kwargs=backend_kwargs,
            type_adapter_callback=type_adapter_callback,
            query_adapter_callback=query_adapter_callback,
            resolve_foreign_keys=resolve_foreign_keys,
            engine_adapter_callback=engine_adapter_callback,
            engine_kwargs=engine_kwargs,
            table_loader_class=table_loader_class,
        )


@decorators.resource(name=lambda args: args["table"], spec=SqlTableResourceConfiguration)
def sql_table(
    credentials: Union[ConnectionStringCredentials, Engine, str] = dlt.secrets.value,
    table: str = dlt.config.value,
    schema: Optional[str] = dlt.config.value,
    metadata: Optional[MetaData] = None,
    incremental: Optional[Incremental[Any]] = None,
    chunk_size: int = 50000,
    backend: TableBackend = "sqlalchemy",
    detect_precision_hints: Optional[bool] = None,
    reflection_level: Optional[ReflectionLevel] = "full",
    defer_table_reflect: Optional[bool] = None,
    table_adapter_callback: Optional[TTableAdapter] = None,
    backend_kwargs: Dict[str, Any] = None,
    type_adapter_callback: Optional[TTypeAdapter] = None,
    included_columns: Optional[List[str]] = None,
    excluded_columns: Optional[List[str]] = None,
    query_adapter_callback: Optional[TQueryAdapter] = None,
    resolve_foreign_keys: bool = False,
    engine_adapter_callback: Callable[[Engine], Engine] = None,
    write_disposition: TWriteDispositionConfig = "append",
    primary_key: TColumnNames = None,
    merge_key: TColumnNames = None,
    engine_kwargs: Optional[Dict[str, Any]] = None,
    table_loader_class: Optional[Type[BaseTableLoader]] = None,
) -> DltResource:
    """
    A dlt resource which loads data from an SQL database table using SQLAlchemy.

    Args:
        credentials (Union[ConnectionStringCredentials, Engine, str]): Database credentials or an `Engine` instance representing the database connection.

        table (str): Name of the table or view to load.

        schema (Optional[str]): Optional name of the schema the table belongs to.

        metadata (Optional[MetaData]): Optional `sqlalchemy.MetaData` instance. If provided, the `schema` argument is ignored.

        incremental (Optional[Incremental[Any]]): Option to enable incremental loading for the table.
            E.g., `incremental=dlt.sources.incremental('updated_at', pendulum.parse('2022-01-01T00:00:00Z'))`

        chunk_size (int): Number of rows yielded in one batch. SQL Alchemy will create additional internal rows buffer twice the chunk size.

        backend (TableBackend): Type of backend to generate table data. One of: "sqlalchemy", "pyarrow", "pandas" and "connectorx".
            "sqlalchemy" yields batches as lists of Python dictionaries, "pyarrow" and "connectorx" yield batches as arrow tables, "pandas" yields pandas DataFrames.
            "sqlalchemy" is the default and does not require additional dependencies, "pyarrow" creates stable destination schemas with correct data types,
            "connectorx" is typically the fastest but ignores the "chunk_size" so you must deal with large tables yourself.

        detect_precision_hints (Optional[bool]): Deprecated. Use `reflection_level`. Set column precision and scale hints for supported data types in the target schema based on the columns in the source tables.
            This is disabled by default.

        reflection_level (Optional[ReflectionLevel]): Specifies how much information should be reflected from the source database schema.
            "minimal": Only table names, nullability and primary keys are reflected. Data types are inferred from the data.
            "full" (default): Data types will be reflected on top of "minimal". `dlt` will coerce the data into reflected types if necessary.
            "full_with_precision": Sets precision and scale on supported data types (ie. decimal, text, binary). Creates big and regular integer types.

        defer_table_reflect (Optional[bool]): Will connect and reflect table schema only when yielding data. Requires `table_names` to be explicitly passed.
            Enable this option when running on Airflow and other orchestrators that create execution DAGs. When True, schema is decided during execution,
            which may override `query_adapter_callback` modifications or `apply_hints`.

        table_adapter_callback (Optional[TTableAdapter]): Receives each reflected table. May be used to modify the list of columns that will be selected.

        backend_kwargs (Dict[str, Any], optional): kwargs passed to table backend ie. "conn" is used to pass specialized connection string to connectorx.

        type_adapter_callback (Optional[TTypeAdapter]): Callable to override type inference when reflecting columns.
            Argument is a single sqlalchemy data type (`TypeEngine` instance) and it should return another sqlalchemy data type, or `None` (type will be inferred from data)

        included_columns (Optional[List[str]]): List of column names to select from the table. If not provided, all columns are loaded.

        excluded_columns (Optional[List[str]]): List of column names to exclude from select. If not provided, all columns are loaded.

        query_adapter_callback (Optional[TQueryAdapter]): Callable to override the SELECT query used to fetch data from the table.
            The callback receives the sqlalchemy `Select` and corresponding `Table`, 'Incremental` and `Engine` objects and should return the modified `Select` or `Text`.

        resolve_foreign_keys (bool): Translate foreign keys in the same schema to `references` table hints.
            May incur additional database calls as all referenced tables are reflected.

        engine_adapter_callback (Callable[[Engine], Engine]): Callback to configure, modify an Engine instance that will be used to open a connection ie. to
            set transaction isolation level.

        write_disposition (TWriteDispositionConfig): write disposition of the table resource, defaults to `append`.
        primary_key (TColumnNames): A list of column names that comprise a private key. Typically used with "merge" write disposition to deduplicate loaded data.
        merge_key (TColumnNames): A list of column names that define a merge key. Typically used with "merge" write disposition to remove overlapping data ranges ie. to
            keep a single record for a given day.

        engine_kwargs (Optional[Dict[str, Any]]): Optional SQLAlchemy engine keyword arguments passed directly to `sqlalchemy.create_engine()`.
            They always affect table reflection and also data loading if SQLAlchemy backend is used (default). If other backend is used, pass equivalent idiomatic params to backend_kwargs.

        table_loader_class (Optional[Type[BaseTableLoader]]): Custom table loader class to use for loading data.
            Subclass `TableLoader` to customize row loading within the SQLAlchemy ecosystem (e.g. pagination,
            retry logic), or subclass `BaseTableLoader` for entirely different backends. When not provided,
            the default loader is selected based on the `backend` parameter.

    Returns:
        DltResource: The dlt resource for loading data from the SQL database table.
    """
    # In case we get None from the config, we want to default to "append"
    write_disposition = write_disposition if write_disposition else "append"

    _detect_precision_hints_deprecated(detect_precision_hints)

    if detect_precision_hints:
        reflection_level = "full_with_precision"
    else:
        reflection_level = reflection_level or "minimal"

    engine_kwargs = engine_kwargs or {}
    engine = engine_from_credentials(
        credentials,
        may_dispose_after_use=True,
        **engine_kwargs,
    )
    engine.execution_options(stream_results=True, max_row_buffer=2 * chunk_size)
    if engine_adapter_callback:
        engine = engine_adapter_callback(engine)
    metadata = metadata or MetaData(schema=schema)
    default_engine_adapter_callback(engine, metadata)

    # Table object is only created when reflecting, we don't want empty tables in metadata
    # as it breaks foreign key resolution
    table_obj = metadata.tables.get(table)
    if table_obj is None and not defer_table_reflect:
        table_obj = Table(table, metadata, autoload_with=engine, resolve_fks=resolve_foreign_keys)

    if table_obj is not None:
        if not defer_table_reflect:
            table_obj = _execute_table_adapter(
                table_obj, table_adapter_callback, included_columns, excluded_columns
            )
        skip_nested_on_minimal = backend == "sqlalchemy"
        hints = table_to_resource_hints(
            table_obj,
            reflection_level,
            type_adapter_callback,
            skip_nested_on_minimal,
            resolve_foreign_keys=resolve_foreign_keys,
        )
    else:
        hints = {}

    if primary_key:
        # may be from what is found in the reflection, so it is set explicitly
        hints["primary_key"] = [primary_key] if isinstance(primary_key, str) else list(primary_key)

    return decorators.resource(
        table_rows,
        name=str(table),
        write_disposition=write_disposition,
        merge_key=merge_key,
        **hints,
    )(
        engine,
        table_obj if table_obj is not None else table,  # Pass table name if reflection deferred
        metadata,
        chunk_size,
        backend,
        incremental=incremental,
        reflection_level=reflection_level,
        table_adapter_callback=table_adapter_callback,
        backend_kwargs=backend_kwargs,
        type_adapter_callback=type_adapter_callback,
        included_columns=included_columns,
        excluded_columns=excluded_columns,
        query_adapter_callback=query_adapter_callback,
        resolve_foreign_keys=resolve_foreign_keys,
        table_loader_class=table_loader_class,
    )


__all__ = [
    "sql_database",
    "sql_table",
    "BaseTableLoader",
    "TableLoader",
    "register_table_loader_backend",
    "get_table_loader_class",
    "ReflectionLevel",
    "TTypeAdapter",
    "engine_from_credentials",
    "remove_nullability_adapter",
    "TableBackend",
    "TQueryAdapter",
    "TTableAdapter",
]



import math
import dataclasses
from base64 import b64encode
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Final,
    Iterable,
    List,
    Literal,
    Optional,
    Union,
    cast,
)
from typing_extensions import Annotated
from requests.auth import AuthBase
from requests import PreparedRequest, Session as BaseSession  # noqa: I251

from dlt.common import logger
from dlt.common.exceptions import MissingDependencyException
from dlt.common.configuration.specs.base_configuration import configspec, NotResolved
from dlt.common.configuration.specs import CredentialsConfiguration
from dlt.common.configuration.specs.exceptions import NativeValueError
from dlt.common.pendulum import pendulum
from dlt.common.typing import TSecretStrValue

if TYPE_CHECKING:
    from cryptography.hazmat.primitives.asymmetric.types import PrivateKeyTypes
else:
    PrivateKeyTypes = Any

TApiKeyLocation = Literal["header", "cookie", "query", "param"]  # Alias for scheme "in" field


class AuthConfigBase(AuthBase, CredentialsConfiguration):
    """Authenticator base which is both `requests` friendly AuthBase and dlt SPEC
    configurable via env variables or toml files
    """

    def __bool__(self) -> bool:
        # This is needed to avoid AuthConfigBase-derived classes
        # which do not implement CredentialsConfiguration interface
        # to be evaluated as False in requests.sessions.Session.prepare_request()
        return True


@configspec
class BearerTokenAuth(AuthConfigBase):
    """Uses `token` for Bearer authentication in "Authorization" header."""

    token: TSecretStrValue = None

    def parse_native_representation(self, value: Any) -> None:
        if isinstance(value, str):
            self.token = value
        else:
            raise NativeValueError(
                type(self),
                value,
                "`token` argument of `BearerTokenAuth` must be a `str`. Got value of type"
                f" `{type(value)}`",
            )

    def __call__(self, request: PreparedRequest) -> PreparedRequest:
        request.headers["Authorization"] = f"Bearer {self.token}"
        return request


@configspec
class APIKeyAuth(AuthConfigBase):
    """Uses provided `api_key` to create authorization data in the specified `location` (query, param, header, cookie) under specified `name`"""

    name: str = "Authorization"
    api_key: TSecretStrValue = None
    location: TApiKeyLocation = "header"

    def parse_native_representation(self, value: Any) -> None:
        if isinstance(value, str):
            self.api_key = value
        else:
            raise NativeValueError(
                type(self),
                value,
                "`api_key` argument of `APIKeyAuth` must be a `str`. Got value of type"
                f" `{type(value)}`",
            )

    def __call__(self, request: PreparedRequest) -> PreparedRequest:
        if self.location == "header":
            request.headers[self.name] = self.api_key
        elif self.location in ["query", "param"]:
            request.prepare_url(request.url, {self.name: self.api_key})
        elif self.location == "cookie":
            raise NotImplementedError()
        return request


@configspec
class HttpBasicAuth(AuthConfigBase):
    """Uses HTTP basic authentication"""

    username: str = None
    password: TSecretStrValue = None

    def parse_native_representation(self, value: Any) -> None:
        if isinstance(value, Iterable) and not isinstance(value, str):
            value = list(value)
            if len(value) == 2:
                self.username, self.password = value
                return
        raise NativeValueError(
            type(self),
            value,
            "`username` and `password` arguments of `HttpBasicAuth` must be a `str` values. "
            f"Got values of type `{type(value)}`",
        )

    def __call__(self, request: PreparedRequest) -> PreparedRequest:
        encoded = b64encode(f"{self.username}:{self.password}".encode()).decode()
        request.headers["Authorization"] = f"Basic {encoded}"
        return request


@configspec
class OAuth2AuthBase(AuthConfigBase):
    """Base class for oauth2 authenticators. requires access_token"""

    access_token: TSecretStrValue = None

    def parse_native_representation(self, value: Any) -> None:
        if isinstance(value, str):
            self.access_token = value
        else:
            raise NativeValueError(
                type(self),
                value,
                "`access_token` argument of `OAuth2AuthBase` must be a `str`. Got value of type"
                f" `{type(value)}`",
            )

    def __call__(self, request: PreparedRequest) -> PreparedRequest:
        request.headers["Authorization"] = f"Bearer {self.access_token}"
        return request


@configspec
class OAuth2ClientCredentials(OAuth2AuthBase):
    """
    This class implements OAuth2 Client Credentials flow where the authorization service
    gives permission without the end user approving.
    This is often used for machine-to-machine authorization.
    The client sends its client ID and client secret to the authorization service which replies
    with a temporary access token.
    With the access token, the client can access resource services.
    """

    access_token: Annotated[Optional[TSecretStrValue], NotResolved()] = None
    access_token_url: str = None
    client_id: TSecretStrValue = None
    client_secret: TSecretStrValue = None
    access_token_request_data: Dict[str, Any] = None
    default_token_expiration: int = 3600
    session: Annotated[BaseSession, NotResolved()] = None

    def __post_init__(self) -> None:
        if self.access_token_request_data is None:
            self.access_token_request_data = {}
        self.token_expiry: pendulum.DateTime = pendulum.now()
        # use default system session unless specified otherwise
        if self.session is None:
            from dlt.sources.helpers import requests

            self.session = requests.client.session

    def __call__(self, request: PreparedRequest) -> PreparedRequest:
        if self.access_token is None or self.is_token_expired():
            self.obtain_token()
        request.headers["Authorization"] = f"Bearer {self.access_token}"
        return request

    def is_token_expired(self) -> bool:
        return pendulum.now() >= self.token_expiry

    def obtain_token(self) -> None:
        response = self.session.post(self.access_token_url, **self.build_access_token_request())
        response.raise_for_status()
        response_json = response.json()
        self.parse_native_representation(self.parse_access_token(response_json))
        expires_in_seconds = self.parse_expiration_in_seconds(response_json)
        self.token_expiry = pendulum.now().add(seconds=expires_in_seconds)

    def build_access_token_request(self) -> Dict[str, Any]:
        return {
            "headers": {
                "Content-Type": "application/x-www-form-urlencoded",
            },
            "data": {
                "client_id": self.client_id,
                "client_secret": self.client_secret,
                "grant_type": "client_credentials",
                **self.access_token_request_data,
            },
        }

    def parse_expiration_in_seconds(self, response_json: Any) -> int:
        return int(response_json.get("expires_in", self.default_token_expiration))

    def parse_access_token(self, response_json: Any) -> str:
        return str(response_json.get("access_token"))


@configspec
class OAuthJWTAuth(BearerTokenAuth):
    """This is a form of Bearer auth, actually there's not standard way to declare it in openAPI"""

    format: Final[Literal["JWT"]] = dataclasses.field(  # noqa: A003
        default="JWT", init=False, repr=False, compare=False
    )
    client_id: str = None
    private_key: TSecretStrValue = None
    auth_endpoint: str = None
    scopes: Optional[Union[str, List[str]]] = None
    headers: Optional[Dict[str, str]] = None
    private_key_passphrase: Optional[TSecretStrValue] = None
    default_token_expiration: int = 3600
    session: Annotated[BaseSession, NotResolved()] = None

    def __post_init__(self) -> None:
        self.scopes = self.scopes if isinstance(self.scopes, str) else " ".join(self.scopes)
        self.token = None
        self.token_expiry: Optional[pendulum.DateTime] = None
        # use default system session unless specified otherwise
        if self.session is None:
            from dlt.sources.helpers import requests

            self.session = requests.client.session

    def __call__(self, r: PreparedRequest) -> PreparedRequest:
        if self.token is None or self.is_token_expired():
            self.obtain_token()
        r.headers["Authorization"] = f"Bearer {self.token}"
        return r

    def is_token_expired(self) -> bool:
        return not self.token_expiry or pendulum.now() >= self.token_expiry

    def obtain_token(self) -> None:
        try:
            import jwt
        except ModuleNotFoundError:
            raise MissingDependencyException("dlt OAuth helpers", ["PyJWT"])

        payload = self.create_jwt_payload()
        data = {
            "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
            "assertion": jwt.encode(payload, self.load_private_key(), algorithm="RS256"),
        }

        logger.debug(f"Obtaining token from {self.auth_endpoint}")

        response = self.session.post(self.auth_endpoint, headers=self.headers, data=data)
        response.raise_for_status()

        token_response = response.json()
        self.token = token_response["access_token"]
        self.token_expiry = pendulum.now().add(
            seconds=token_response.get("expires_in", self.default_token_expiration)
        )

    def create_jwt_payload(self) -> Dict[str, Union[str, int]]:
        now = pendulum.now()
        return {
            "iss": self.client_id,
            "sub": self.client_id,
            "aud": self.auth_endpoint,
            "exp": math.floor((now.add(hours=1)).timestamp()),
            "iat": math.floor(now.timestamp()),
            "scope": cast(str, self.scopes),
        }

    def load_private_key(self) -> "PrivateKeyTypes":
        try:
            from cryptography.hazmat.backends import default_backend
            from cryptography.hazmat.primitives import serialization
        except ModuleNotFoundError:
            raise MissingDependencyException("dlt OAuth helpers", ["cryptography"])

        private_key_bytes = self.private_key.encode("utf-8")
        return serialization.load_pem_private_key(
            private_key_bytes,
            password=(
                self.private_key_passphrase.encode("utf-8") if self.private_key_passphrase else None
            ),
            backend=default_backend(),
        )



"""String formatting, datetime humanization, and dict-to-table conversion utilities for the dashboard."""

from datetime import datetime  # noqa: I251
from itertools import chain
from typing import Any, Dict, List, Mapping, Union, cast

from dlt._workspace.helpers.dashboard.typing import TNameValueItem

from dlt.common.pendulum import pendulum

from dlt.common.configuration.exceptions import ConfigFieldMissingException
from dlt.common.destination.exceptions import SqlClientNotAvailable
from dlt.destinations.exceptions import DatabaseUndefinedRelation, DestinationUndefinedEntity
from dlt.pipeline.exceptions import PipelineConfigMissing

from dlt._workspace.helpers.dashboard import strings
from dlt._workspace.helpers.dashboard.config import DashboardConfiguration


def humanize_datetime(dt: Union[str, int, float, datetime], datetime_format: str) -> str:
    """Format a datetime-like value (datetime, timestamp, or ISO string) into a human-readable string."""
    if dt in ["", None, "-"]:
        return "-"
    elif isinstance(dt, datetime):
        p = pendulum.instance(dt)
    elif isinstance(dt, str) and dt.replace(".", "").isdigit():
        p = pendulum.from_timestamp(float(dt))
    elif isinstance(dt, str):
        p = cast(pendulum.DateTime, pendulum.parse(dt))
    elif isinstance(dt, (int, float)):
        p = pendulum.from_timestamp(dt)
    else:
        raise ValueError(f"Invalid datetime value: {dt}")
    return p.format(datetime_format)


def humanize_datetime_values(c: DashboardConfiguration, d: Dict[str, Any]) -> Dict[str, Any]:
    """Format well-known datetime keys (started_at, finished_at, etc.) in a dict to human-readable strings."""
    _DATETIME_KEYS = ("started_at", "finished_at", "inserted_at", "created", "last_modified")

    # compute duration before we overwrite the raw values
    started_at = d.get("started_at", "")
    finished_at = d.get("finished_at", "")
    if started_at not in ("", None, "-") and finished_at not in ("", None, "-"):
        d["duration"] = (
            pendulum.instance(finished_at).diff(pendulum.instance(started_at)).in_words()
        )

    for key in _DATETIME_KEYS:
        if d.get(key):
            d[key] = humanize_datetime(d[key], c.datetime_format)

    # load_id is a timestamp that maps to a separate display key
    if d.get("load_id"):
        d["load_package_created_at"] = humanize_datetime(d["load_id"], c.datetime_format)

    return d


def format_duration(ms: float) -> str:
    """Format a duration in milliseconds as a human-readable string (e.g. '120ms', '3.5s', '1.2m')."""
    if ms < 1000:
        return f"{int(ms)}ms"
    elif ms < 60000:
        return f"{round(ms / 100) / 10}s"
    else:
        return f"{round(ms / 6000) / 10}m"


def format_exception_message(exception: Exception) -> str:
    """Convert a pipeline/destination exception to a user-friendly error message."""
    if isinstance(exception, (PipelineConfigMissing, ConfigFieldMissingException)):
        return strings.error_config_missing
    elif isinstance(exception, (SqlClientNotAvailable)):
        return strings.error_sql_not_supported
    elif isinstance(exception, (DestinationUndefinedEntity, DatabaseUndefinedRelation)):
        return strings.error_undefined_entity
    return str(exception)


def filter_empty_values(d: Mapping[str, Any]) -> Dict[str, Any]:
    """Return a new dict with all None and empty string values removed."""
    return {k: v for k, v in d.items() if v is not None and v != ""}


def align_dict_keys(items: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """Ensure all dicts in a list share the same keys, filling missing ones with '-' for uniform table rendering."""
    items = cast(List[Dict[str, Any]], [filter_empty_values(i) for i in items])
    all_keys = set(chain.from_iterable(i.keys() for i in items))
    for i in items:
        i.update({key: "-" for key in all_keys if key not in i})
    return items


def dict_to_table_items(d: Dict[str, Any]) -> List[TNameValueItem]:
    """Convert a dict to a list of {name, value} dicts for display in a marimo table."""
    return [{"name": k, "value": v} for k, v in d.items()]



from dlt.common.configuration.specs import (
    GcpServiceAccountCredentials,
    GcpOAuthCredentials,
    GcpCredentials,
    AwsCredentials,
    AzureCredentials,
    AzureServicePrincipalCredentials,
    SFTPCredentials,
)
from dlt.common.configuration.specs import ConnectionStringCredentials
from dlt.common.configuration.specs import OAuth2Credentials
from dlt.common.configuration.specs import CredentialsConfiguration, configspec
from dlt.common.storages.configuration import (
    FileSystemCredentials,
    FilesystemConfiguration,
    FilesystemConfigurationWithLocalFiles,
)


__all__ = [
    "GcpServiceAccountCredentials",
    "GcpOAuthCredentials",
    "GcpCredentials",
    "AwsCredentials",
    "AzureCredentials",
    "ConnectionStringCredentials",
    "OAuth2Credentials",
    "CredentialsConfiguration",
    "configspec",
    "FileSystemCredentials",
    "FilesystemConfiguration",
    "FilesystemConfigurationWithLocalFiles",
    "AzureServicePrincipalCredentials",
    "SFTPCredentials",
]



from typing import Dict, Optional

from dlt.common.schema.typing import TColumnName, TypedDict


class RelationalNormalizerConfigPropagation(TypedDict, total=False):
    root: Optional[Dict[TColumnName, TColumnName]]
    tables: Optional[Dict[str, Dict[TColumnName, TColumnName]]]


class RelationalNormalizerConfig(TypedDict, total=False):
    max_nesting: Optional[int]
    propagation: Optional[RelationalNormalizerConfigPropagation]
    root_key_propagation: Optional[bool]



from typing import Optional, Dict, Iterator, Any

from dlt.common import jsonpath

from .client import RESTClient  # noqa: F401
from .client import PageData
from .auth import AuthConfigBase
from .paginators import BasePaginator
from .typing import HTTPMethodBasic, Hooks


def paginate(
    url: str,
    method: HTTPMethodBasic = "GET",
    headers: Optional[Dict[str, str]] = None,
    params: Optional[Dict[str, Any]] = None,
    json: Optional[Dict[str, Any]] = None,
    auth: AuthConfigBase = None,
    paginator: Optional[BasePaginator] = None,
    data_selector: Optional[jsonpath.TJsonPath] = None,
    hooks: Optional[Hooks] = None,
    *,
    data: Optional[Any] = None,
) -> Iterator[PageData[Any]]:
    """
    Paginate over a REST API endpoint.

    Args:
        url: URL to paginate over.
        **kwargs: Keyword arguments to pass to `RESTClient.paginate`.

    Returns:
        Iterator[Page]: Iterator over pages.
    """
    client = RESTClient(
        base_url=url,
        headers=headers,
    )
    return client.paginate(
        path="",
        method=method,
        params=params,
        json=json,
        auth=auth,
        paginator=paginator,
        data_selector=data_selector,
        hooks=hooks,
        data=data,
    )
